/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

import java.awt.*;
import javax.swing.*;

/**
 *  Description of the Class
 *
 * @author     rsheh
 * @created    19 December 2002
 */
public class BFLPanel extends JPanel
{
    public final static byte ORANGE         = 0;
    public final static byte BLUE           = 1;
    public final static byte GREEN          = 2;
    public final static byte YELLOW         = 3;
    public final static byte PINK           = 4;
    public final static byte BLUEDOG        = 5;
    public final static byte REDDOG         = 6;
    public final static byte GREENFIELD     = 7;
    public final static byte ROBOTGREY      = 8;
    public final static byte FIELDWHITE     = 9;
    public final static byte BLACK          = 10;
    public final static byte NONE           = 127;

	public               byte      color        = NONE;
	//the color we are higlighting (9 is none)
	public               boolean   drawOverlay  = true;
	//arrays for colour data
	                     byte[][]  Y            = new byte[PIC_WIDTH][PIC_HEIGHT];
	                     byte[][]  U            = new byte[PIC_WIDTH][PIC_HEIGHT];
	                     byte[][]  V            = new byte[PIC_WIDTH][PIC_HEIGHT];
	                     byte[][]  C            = new byte[PIC_WIDTH][PIC_HEIGHT];
											 byte[][]  prevC        = new byte[PIC_WIDTH][PIC_HEIGHT];

	final static         int       PIC_WIDTH    = 176;
	final static         int       PIC_HEIGHT   = 144;

	                     int       SCALE        = 1;
	                     int       OLDSCALE     = -1;
	                     Image     bi;
	                     Graphics  big;

	public               int[][]   rgbPlane     = new int[PIC_WIDTH][PIC_HEIGHT];
	public               int[][]   rgbOverlay   = new int[PIC_WIDTH][PIC_HEIGHT];

	                     int       STARTX       = 0;
	                     int       STARTY       = 0;

	                     int       ENDX         = PIC_WIDTH;
	                     int       ENDY         = PIC_HEIGHT;

	                     int       quadrant     = 0;


	/**  Constructor for the BFLPanel object */
	public BFLPanel()
	{
		super();
		for (int h = 0; h < PIC_HEIGHT; h++)
		{
			for (int w = 0; w < PIC_WIDTH; w++)
			{
				rgbOverlay[w][h] = -1;
			}
		}
	}


	/**
	 *  Description of the Method
	 *
	 * @param  g  Description of the Parameter
	 */
	public void paintComponent(Graphics g)
	{

		SCALE = getWidth() / PIC_WIDTH;
		int  SCALE2  = getHeight() / PIC_HEIGHT;

		if (SCALE2 < SCALE)
		{
			SCALE = SCALE2;
		}

		//recreate offscreen buffers only if the scale has changed
		if (SCALE != OLDSCALE)
		{
			bi = createImage(getWidth(), getHeight());
			big = bi.getGraphics();
			OLDSCALE = SCALE;
		}

		//set scale twice as big if we are zoomed on a quadrant
		if (quadrant != 0)
		{
			SCALE *= 2;
		}

		for (int x = STARTX; x < ENDX; x++)
		{
			for (int y = STARTY; y < ENDY; y++)
			{

				int  y1  = Y[x][y] & 0xff;
				int  u   = U[x][y] & 0xff;
				int  v   = V[x][y] & 0xff;

				drawSquare(x, y, big);

				//if pixel is classified draw marking box
				if (color != NONE && C[x][y] == color)
				{
					drawBox(x, y, big);
				}
			}
		}

		//draw buffered image to screen
		g.drawImage(bi, 0, 0, this);
	}


	/**
	 *  Description of the Method
	 *
	 * @param  x  Description of the Parameter
	 * @return    Description of the Return Value
	 */
	public int convertX(int x)
	{

		int  r;

		//divide by scale and add offset
		r = x / SCALE + STARTX;

		return r;
	}


	/**
	 *  Description of the Method
	 *
	 * @param  y  Description of the Parameter
	 * @return    Description of the Return Value
	 */
	public int convertY(int y)
	{

		int  r;

		//divide by scale and add offset
		r = y / SCALE + STARTY;

		return r;
	}


	//draw a pixel square with black border
	/**
	 *  Description of the Method
	 *
	 * @param  x     Description of the Parameter
	 * @param  y     Description of the Parameter
	 * @param  picg  Description of the Parameter
	 */
	public void drawSquare(int x, int y, Graphics picg)
	{

		int  xd  = x - STARTX;
		int  yd  = y - STARTY;

		//draw pixel in correct rgb color
		//draw rest of pixel in correct rgb
		picg.setColor(new Color(rgbPlane[x][y]));
		picg.fillRect(xd * SCALE, yd * SCALE, SCALE, SCALE);
		//draw black border
		//picg.setColor(Color.black);
		picg.drawRect(xd * SCALE, yd * SCALE, SCALE, SCALE);

		if ((rgbOverlay[x][y] != -1) && drawOverlay)
		{
			picg.setColor(new Color(rgbOverlay[x][y]));
//		picg.fillRect(xd*SCALE,yd*SCALE,SCALE,SCALE);
//		picg.drawRect(xd*SCALE,yd*SCALE,SCALE,SCALE);
			picg.drawLine(xd * SCALE, yd * SCALE + SCALE / 2, (xd + 1) * SCALE, yd * SCALE + SCALE / 2);
//		picg.fillOval(xd*SCALE, yd*SCALE, SCALE, SCALE);
			picg.setColor(new Color(rgbPlane[x][y]));
		}

	}


	//draw magenta box indicating classified
	/**
	 *  Description of the Method
	 *
	 * @param  x     Description of the Parameter
	 * @param  y     Description of the Parameter
	 * @param  picg  Description of the Parameter
	 */
	public void drawBox(int x, int y, Graphics picg)
	{

		int  xd  = x - STARTX;
		int  yd  = y - STARTY;
/*
		if (this.color == 4 || this.color == 6)
		{
			picg.setColor(Color.cyan);
		}
		else
		{
			picg.setColor(Color.black);
		}
*/
//		picg.setColor((new Color(rgbPlane[x][y] ^ 0x0FFFFF)));
		float hue = (float) 0.5 + 
			Color.RGBtoHSB(	(rgbPlane[x][y] >>> 16) & 0x0FF, 
											(rgbPlane[x][y] >>> 8) & 0x0FF,
											 rgbPlane[x][y]         & 0x0FF, null)[0];
		if (hue > (float)1.0)
		{
			hue = hue - (float)1.0;
		}
		picg.setColor(Color.getHSBColor(hue, (float)1.0, (float)1.0));
		picg.drawRect(xd * SCALE + 1, yd * SCALE + 1, SCALE - 2, SCALE - 2);
	}


	//set the quadrant to zoom on (0 is all quadrants)
	/**
	 *  Sets the quadrant attribute of the BFLPanel object
	 *
	 * @param  x  The new quadrant value
	 * @param  y  The new quadrant value
	 */
	public void setQuadrant(int x, int y)
	{

		if (x == -1 && y == -1)
		{
			quadrant = 0;
			STARTX = 0;
			STARTY = 0;
			ENDX = PIC_WIDTH;
			ENDY = PIC_HEIGHT;
		}

		else
		{
			quadrant = 1;
			STARTX = Math.max(x - PIC_WIDTH / 4, 0);
			STARTY = Math.max(y - PIC_HEIGHT / 4, 0);
			ENDX = STARTX + PIC_WIDTH / 2;
			ENDY = STARTY + PIC_HEIGHT / 2;

			//correct out of bounds zooming
			if (ENDX > PIC_WIDTH)
			{
				STARTX = PIC_WIDTH / 2;
				ENDX = PIC_WIDTH;
			}
			if (ENDY > PIC_HEIGHT)
			{
				STARTY = PIC_HEIGHT / 2;
				ENDY = PIC_HEIGHT;
			}
		}

		//update graphics when set quadrant is called
		paintComponent(getGraphics());

	}

	public void saveUndo()
	{
		for (int h = 0; h < PIC_HEIGHT; h ++)
		{
			for (int w = 0; w < PIC_WIDTH; w ++)
			{
				prevC[w][h] = C[w][h];
			}
		}
	}

	public void undo()
	{
		for (int h = 0; h < PIC_HEIGHT; h ++)
		{
			for (int w = 0; w < PIC_WIDTH; w ++)
			{
				C[w][h] = prevC[w][h];
			}
		}
		this.repaint();
	}


}

