/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "viewer.h"


#define TRUE 1
#define FALSE -1
#define R_DEG 57.29              // degrees per radian
#define BLACK 10

static FLIMAGE_SETUP setup;      // set up param
int V_WIDTH           = 320;
int V_HEIGHT          = 240;
int PORT              = 1;       // Composite
int NORM              = 0;       // Pal = 0  NTSC=1
int MAX_VAL           = 255;     // *shrug*
int MAX_OBJ           = 12;      // maximum no of each object
int SPOT_D            = 4;
int BALL_D            = 6;
int DOG_D             = 8;
int COLOUR_CUBE_SIZE  = 128;     // size of cube generated from ic

int running = FALSE;
int display = FALSE;
int fd;                          // frame grabber id
int *undistort;                  // distort - correct pixel mapping
char *colourcube;                // calibrated colours
FD_main *main_form;              // the main form
FL_IMAGE *image; 
char *cplane;
List run;
Point *red_spot, *blue_spot;     // list of balls
Point *ball_spot;                     // size of each thing
int blue_count, red_count, ball_count;
FixedPoint *fixed_point;         // for determining image -> world coordinate transform
float pixel_size;                // distance one pixel represents
int fixed_point_length;
Mobile dog, ball;

void play_video();
void init_calib(char *filename1, char *filename2);
void init_space(char *filename);
void init_forms();
void init_video();

void process_image(unsigned char *frame);
void identify_spots();
void identify_objects();
void colour_image();
void colour_run();
void colour_blob();
void colour_dog();
char colour_check(int r, int g, int b);
void colour_object(int row, int col, int run_colour);
void colour_pixel(int row, int col, int r, int g, int b);
void draw_cross(int midy, int midx, int colour, int ra);
int normalise(int x1, int y1, int x2, int y2);
Point world_coordinate(int x, int y);
float spot_distance(Point p1, Point p2);
void add_spot(int x, int y, int c);
void update_mobile(Mobile *m, Point p, Point d, int angle);

/*---------------------------------------
 * generated from fdesign
 *---------------------------------------*/

FD_main *create_form_main(void) {
  FL_OBJECT *obj;
  FD_main *fdui = (FD_main *) fl_calloc(1, sizeof(*fdui));

  fdui->main = fl_bgn_form(FL_NO_BOX, V_WIDTH+1, V_HEIGHT+191);
  obj = fl_add_box(FL_RFLAT_BOX,0,0,V_WIDTH+1,V_HEIGHT+191,"");
  obj = fl_add_canvas(FL_NORMAL_CANVAS,0,0,V_WIDTH,V_HEIGHT,"");
  fdui->canvas = obj;
  obj = fl_add_button(FL_NORMAL_BUTTON,30,V_HEIGHT,60,30,"go");
    fl_set_object_callback(obj,go_pressed,0);
  obj = fl_add_button(FL_NORMAL_BUTTON,V_WIDTH-90,V_HEIGHT,60,30,"exit");
    fl_set_object_callback(obj,exit_pressed,0);
  obj = fl_add_text(FL_NORMAL_TEXT,0,V_HEIGHT+30,V_WIDTH,145,"Info ..");
    fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
  fdui->info = obj;
  obj = fl_add_text(FL_NORMAL_TEXT,0,V_HEIGHT+175,V_WIDTH,15,"Frame Rate ..");
    fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
  fdui->frate = obj;
  fl_end_form();

  fdui->main->fdui = fdui;

  return fdui;
}

/*---------------------------------------
 * Main 
 *---------------------------------------*/

int main(int argc, char *argv[]) {
  if (argc <= 3) {
    printf("Usage: viewer <radial correction file> <colour calibration file> <spatial calibration file> <optional>\n");
    exit(1);
  }
  if (argc > 4) display = TRUE;
  fl_initialize(&argc, argv, 0, 0, 0);
  init_calib(argv[1], argv[2]);
  init_space(argv[3]);
  init_forms();
  init_video();

  //init_blog();

  fl_do_forms();
  return 0;
}

/*---------------------------------------
 * Button handlers
 *---------------------------------------*/

// When go is pressed, init if not already done so
// and display images captured from camera
void go_pressed(FL_OBJECT *ob, long data) {
  if (running == FALSE) {
    running = TRUE;
    fl_set_object_label(ob, "Stop");
    play_video();
  } else {
    running = FALSE;
    fl_set_object_label(ob, "Go");
  }
}

// clean up and exit
void exit_pressed(FL_OBJECT *ob, long data) {
  if (running == TRUE) {
    close_video(fd);
    free(undistort);
    free(red_spot);
    free(blue_spot);
    free(ball_spot);
    free(colourcube);
    free(fixed_point);
    //close_blog();
    if (image != NULL)
      flimage_free(image);
    if (cplane != NULL)
      free(cplane);
  }
  exit(0);
}

/*---------------------------------------
 * Initialisation
 *---------------------------------------*/

// read calibration file
void init_calib(char *filename1, char *filename2) {
  FILE *fp;
  fp = fopen(filename1, "r");
  if (fp == NULL) {
    printf("Correction file not found\n");
    exit(1);
  }
  undistort = calloc(V_HEIGHT*V_WIDTH*2, sizeof(int));
  fread(undistort, sizeof(int), V_HEIGHT*V_WIDTH*2, fp);
  fclose(fp);
  printf("Done reading radial distortion calibration file.\n");

  fp = fopen(filename2, "r");
  if (fp == NULL) {
    printf("Colour file not found\n");
    exit(1);
  }
  colourcube = calloc(COLOUR_CUBE_SIZE*COLOUR_CUBE_SIZE*COLOUR_CUBE_SIZE, sizeof(char));
  fread(colourcube, sizeof(char), COLOUR_CUBE_SIZE*COLOUR_CUBE_SIZE*COLOUR_CUBE_SIZE, fp);
  fclose(fp);
  printf("Done reading colour calibration file.\n");
}

// init form
void init_forms() {
  red_spot = calloc(MAX_OBJ, sizeof(Point));
  blue_spot = calloc(MAX_OBJ, sizeof(Point));
  ball_spot = calloc(MAX_OBJ, sizeof(Point));
  cplane = calloc(V_HEIGHT * V_WIDTH, sizeof(char));
  main_form = create_form_main();
  setup.app_data = main_form;  // dont know what these are, but neccesary
  setup.delay = 70;    // same as above, from Phil's code
  flimage_setup(&setup);
  flimage_enable_gif();
  fl_show_form(main_form->main, FL_PLACE_FREE,FL_FULLBORDER,"Overhead Camera Viewer");
  printf("Done initialising forms.\n");
}  

// init video
void init_video() {
  fd = open_video("/dev/video0", PORT, NORM, V_WIDTH, V_HEIGHT);
  if( fd == 0) {
    printf("Initialisation failed on port %d\n", PORT);
    exit(1);
  }
  printf("Done initialising video, fd=%d\n", fd);
}

// init pixel to world map
void init_space(char *filename) {
  FILE *fp;
  int i, x, y;
  float xf, yf;
  fp = fopen(filename, "r");
  if (fp == NULL) {
    printf("Spatial file not found\n");
    exit(1);
  }
  fscanf(fp, "%d %f", &fixed_point_length, &pixel_size);
  fixed_point = calloc(fixed_point_length, sizeof(FixedPoint));
  for (i = 0; i < fixed_point_length; i++) {
    fscanf(fp, " %d %d %f %f", &x, &y, &xf, &yf);
    fixed_point[i].p.x = x;
    fixed_point[i].p.y = y;
    fixed_point[i].x = xf;
    fixed_point[i].y = yf;
  }
  printf("Done reading spatial file\nread %d points, pixel = %f cm\n", fixed_point_length, pixel_size);
  fclose(fp);
}

/*---------------------------------------
 * Image Processing
 *---------------------------------------*/

// keep grabbing data and displaying until someone says stop
void play_video() {
  struct timeb t1, t2;
  double       ut1, ut2;
  int frames;
  char l[100];

  unsigned char *frame;
  image = flimage_alloc();
  
  frames = 0;
  DTIME(t1, ut1);
  while (running == TRUE) {
    frames++;
    fl_check_forms();  // i dont know.. maybe
    frame = get_frame(fd);

    if (display == TRUE) {
      flimage_free(image);  // free the last image
      image = flimage_alloc(); // set up for current image

      image->type = FL_IMAGE_RGB;
      image->w = V_WIDTH;
      image->h = V_HEIGHT;
      image->map_len = MAX_VAL+1;  // what ever this means, something like maxVal maybe
      flimage_getmem(image);
    }

    process_image(frame);

    if (display == TRUE) {
      //colour_image(); // displays cplane, debug
      //colour_run(); // displays run, debug
      //colour_blob(); // display blob, debug
      colour_dog();
      flimage_sdisplay(image, FL_ObjWin(main_form->canvas));
    }
    
    if (frames == 10) {
      DTIME(t2, ut2);
      sprintf(l, "%5.2f fps", frames/(ut2 - ut1));
      fl_set_object_label(main_form->frate, l);
      frames = 0;
      DTIME(t1, ut1);
    }
  }
}

// remove the radial distortion
void process_image(unsigned char *frame) {
  int i, j, p, px, py, f, r, g, b, c, cp, lastc;
  Run current;
  p = f = cp = 0;
  initialize_list(&run);
  for (i = 0; i < V_HEIGHT; i++) {
    lastc = c = BACKGROUND;
    for (j = 0; j < V_WIDTH; j++) {
      px = undistort[p++];
      py = undistort[p++];
      if (px >= 0 && px < V_WIDTH && py >= 0 && py < V_HEIGHT) {
      
        f = (px + py * V_WIDTH) * 3;
        b = frame[f++];
        g = frame[f++];
        r = frame[f++];

        if (run_check(r, g, b, lastc) == FALSE) {
          c = colour_check(r, g, b);
	  if (lastc == BACKGROUND || lastc == GREEN_FIELD) {
	    if (c == BACKGROUND || c == GREEN_FIELD) {
	      // do nothing;
	    } else {
	      current.start.x = j;
	      current.start.y = i;
	      current.colour = c;
	      insert_after_cur(&run, current);
	      goto_right(&run);
	    }
	  } else {
	    if (c == BACKGROUND || c == GREEN_FIELD) {
	      (run.current_pos)->data.end.x = j - 1;
	      (run.current_pos)->data.end.y = i;
	    } else {
	      (run.current_pos)->data.end.x = j - 1;
	      (run.current_pos)->data.end.y = i;
	      current.start.x = j;
	      current.start.y = i;
	      current.colour = c;
	      insert_after_cur(&run, current);
	      goto_right(&run);
            }
	  }	  
          lastc = c;
        }
	if (display == TRUE) {
          colour_pixel(i, j, r, g, b);
        }
        cplane[cp++] = c;
      } else {
	if (display == TRUE) {
          colour_pixel(i, j, 0, 0, 0);
        }
        cplane[cp++] = BACKGROUND;
	if (lastc != BACKGROUND && lastc != GREEN_FIELD) {
	  (run.current_pos)->data.end.x = j - 1;
	  (run.current_pos)->data.end.y = i;
          lastc = BACKGROUND;
	}
      }
    }
    if (lastc != BACKGROUND && lastc != GREEN_FIELD) {
      (run.current_pos)->data.end.x = j - 1;
      (run.current_pos)->data.end.y = i;
    }
  }
  identify_spots();
  identify_objects();
}

void identify_spots() {
  int n, s, w, e, maxe, minw, c;
  int n2, s2, w2, e2, c2, sum, d;
  Run r;
  goto_head(&run);
  goto_right(&run);
  red_count = blue_count = ball_count = 0;
  while (run.current_pos != run.head && run.current_pos != NULL) {
    r = (run.current_pos)->data; 
    n = s = r.start.y;
    w = minw = r.start.x;
    e = maxe = r.end.x;
    c = r.colour;
    sum = (w+e)/2;
    delete_at_current(&run);
    while (run.current_pos != run.head && run.current_pos != NULL) {
      r = (run.current_pos)->data;
      n2 = s2 = r.start.y;
      w2 = r.start.x;
      e2 = r.end.x;
      c2 = r.colour;
      if (n2 > s + 1) {
        if (c == BALL)
	  d = BALL_D;
	else
	  d = SPOT_D; 
	if (maxe-minw+1 > SPOT_D && s-n+1 > d) {
  	  add_spot(sum/(s-n+1) + 1, (s+n)/2 + 1, c);
	}
	break;
      } else {
        if (c2 == c) {
	  if (n2 == s && w2 <= e + 1) {
	    sum -= (w + e)/2;
	    sum += (w + e2)/2;
	    e = e2;
	    maxe = max(e, maxe);
	    delete_at_current(&run);
	  } else if (n2 == s+1 && w2 < e && e2 > w) {
	    e = e2;
	    w = w2;
	    maxe = max(e, maxe);
	    minw = min(w, minw);
	    sum += (w2 + e2)/2;
	    s = n2;
	    delete_at_current(&run);	    
	  } else {
  	    goto_right(&run);
	  }
	} else {
	  goto_right(&run);
	}
      }
    }
    goto_head(&run);
    goto_right(&run);
  }
  if (maxe-minw > SPOT_D && s-n > SPOT_D)
    add_spot(sum/(s-n+1) + 1, (s+n)/2 + 1, c);
}

// temporary implementation, was originally designed to cope with multiple balls and robots
// currently assume one ball and one robot
void identify_objects() {  
  int i, j, found, angle, x, y;
  Point w, p, d;
  struct timeb tb;
  double ut;
  char l[300], t[60];
  sprintf(l, "");
  sprintf(t, "");

  // first find a pair of red and blue spot that is most likely
  found = FALSE;
  if (red_count > 0 && blue_count > 0) {
    for (i = 0; i < blue_count; i++) {
      for (j = 0; j < red_count; j++) {
        if (spot_distance(red_spot[j], blue_spot[i]) < 20) {
	  found = TRUE;
	  break;
	}
      }
      if (found == TRUE) break;
    }
    if (found == TRUE) {
      x = (red_spot[j].x + blue_spot[i].x)/2;
      y = (red_spot[j].y + blue_spot[i].y)/2;
      w = world_coordinate(x, y);
      d.x = x; d.y = y;
      angle = (90 + normalise(red_spot[j].x, V_HEIGHT-red_spot[j].y, blue_spot[i].x, V_HEIGHT-blue_spot[i].y)) % 360;
      update_mobile(&dog, w, d, angle);
      p = dog.p;
      sprintf(t, "Dog at.. %d, %d (cm), heading %d (deg)\n", p.x, p.y, dog.angle);
    }
  } else {
    sprintf(t, "No Dog\n");
    dog.p.x = dog.p.y = dog.d.x = dog.d.y = 0;
  }
  strcat(l, t);
  
  if (ball_count > 0) {
    w = world_coordinate(ball_spot[0].x, ball_spot[0].y);
    update_mobile(&ball, w, ball_spot[0], 0);
    p = ball.p;
    sprintf(t, "Ball at.. %d, %d (cm)\n", p.x, p.y);
    DTIME(tb, ut);
    printf("Ball at.. %d, %d (cm) time %g\n", p.x, p.y, ut);
  } else {
    sprintf(t, "No Ball\n");
    ball.p.x = ball.p.y = ball.d.x = ball.d.y = 0;
  }
  strcat(l, t);
  
  fl_set_object_label(main_form->info, l);
}

void colour_image() {
  int i, j, cp;
  cp = 0;
  for (i = 0; i < V_HEIGHT; i++) {
    for (j = 0; j < V_WIDTH; j++) {
      colour_object(i, j, cplane[cp++]);
    }
  }
}

void colour_run() {
  int sx, ex, y, i, c;
  Run r;
  goto_head(&run);
  goto_right(&run);
  while(run.current_pos != run.head && run.current_pos != NULL) {
    r = (run.current_pos)->data;
    sx = r.start.x;
    ex = r.end.x;
    y = r.start.y;
    c = r.colour;
    for (i = sx; i < ex; i++) {
      colour_object(y, i, c);
    }
    goto_right(&run);
  }
}

void colour_blob() {
  int i;
  for (i = 0; i < red_count; i++) {
    draw_cross(red_spot[i].y, red_spot[i].x, BLACK, SPOT_D);
  }
  for (i = 0; i < blue_count; i++) {
    draw_cross(blue_spot[i].y, blue_spot[i].x, BLACK, SPOT_D);
  }
  for (i = 0; i < ball_count; i++) {
    draw_cross(ball_spot[i].y, ball_spot[i].x, BLACK, SPOT_D);
  }
}

void colour_dog() {
  struct timeb t;
  double t1;
  DTIME(t, t1);
  //printf("dog %d %d\n", dog.d.x, dog.d.y);
  draw_cross(dog.d.y, dog.d.x, BLACK, DOG_D);
  draw_cross(ball.d.y, ball.d.x, BLACK, BALL_D);
}
/*---------------------------------------
 * Utility functions
 *---------------------------------------*/

int run_check(int r, int g, int b, int colour) {
  if (r >= 250 && g >= 250 && b >= 250)
    return FALSE;
  if (colour == RED_DOG) {
    if (r >= b && b > g && r > 200) return TRUE;
  } else if (colour == BLUE_DOG) {
    if (b >= r && r > g && b > 200) return TRUE;
  } else if (colour == BALL) {
    if (r >= g && g > b && r > 200) return TRUE;
  }
  return FALSE;
}

char colour_check(int r, int g, int b) {
  char c = colourcube[(r/2) * COLOUR_CUBE_SIZE * COLOUR_CUBE_SIZE + (g/2) * COLOUR_CUBE_SIZE + (b/2)];
  return c;
}

void colour_object(int row, int col, int colour) {
  if (colour == RED_DOG) {
    colour_pixel(row, col, MAX_VAL, 0, 0);
  } else if (colour == BLUE_DOG) {
    colour_pixel(row, col, 0, 0, MAX_VAL);
  } else if (colour == BALL) {
    colour_pixel(row, col, MAX_VAL, MAX_VAL, 0);
  } else if (colour == GREEN_FIELD) {
    colour_pixel(row, col, 0, MAX_VAL, 0);
  }
}

void colour_pixel(int row, int col, int r, int g, int b) {
  image->red[row][col] = r;
  image->green[row][col] = g;
  image->blue[row][col] = b;
}

void draw_cross(int midy, int midx, int colour, int ra) {
  int i, r, g, b;
  int n, s, e, w;
    
  n = midy - ra;
  s = midy + ra;
  w = midx - ra;
  e = midx + ra;
  if (n < 0) n = 0;
  if (s >= V_HEIGHT) s = V_HEIGHT - 1;
  if (w < 0) w = 0;
  if (e >= V_WIDTH) e = V_WIDTH - 1;
  
  if (colour == RED_DOG) {
    r = MAX_VAL; g = 0; b = 0;
  } else if (colour == BLUE_DOG) {
    r = 0; g = 0; b = MAX_VAL;
  } else if (colour == BALL) {
    r = MAX_VAL; g = MAX_VAL; b = 0;
  } else {
    r = g = b = 0;
  }
  
  for (i = n; i < s; i++) {
    image->red[i][midx] = r;
    image->blue[i][midx] = b;
    image->green[i][midx] = g;
  }
  if (midx - ra >= 0) {
    for (i = n; i < s; i++) {
      image->red[i][w] = r;
      image->blue[i][w] = b;
      image->green[i][w] = g;
    }
  }
  if (midx + ra < V_WIDTH) {
    for (i = n; i < s; i++) {
      image->red[i][e] = r;
      image->blue[i][e] = b;
      image->green[i][e] = g;
    }
  }

  for (i = w; i < e; i++) {
    image->red[midy][i] = r;
    image->blue[midy][i] = b;
    image->green[midy][i] = g;
  }
  if (midy - ra >= 0) {
    for (i = w; i < e; i++) {
      image->red[n][i] = r;
      image->blue[n][i] = b;
      image->green[n][i] = g;
    }
  }
  if (midy + ra < V_HEIGHT) {
    for (i = w; i < e; i++) {
      image->red[s][i] = r;
      image->blue[s][i] = b;
      image->green[s][i] = g;
    }
  }
}

int normalise(int x1, int y1, int x2, int y2) {
  int dx, dy;
  float angle;
  dx = x1 - x2;
  dy = y1 - y2;
  angle = atan(1.0f*dy/dx) * R_DEG;
  if (dx >= 0 && dy >= 0)
    return floor(angle);
  else if (dx >= 0 && dy < 0)
    return floor(360 + angle);
  else
    return floor(180 + angle);
}

Point world_coordinate(int x, int y) {
  Point p;
  FixedPoint f;
  int i, min_i;
  float d, min_d;
  min_i = 0;
  min_d = 160000;
  for (i = 0; i < fixed_point_length; i++) {
    f = fixed_point[i];
    d = ((f.p.x-x) * (f.p.x-x) + (f.p.y-y) * (f.p.y-y));
    if (d < min_d) {
      min_d = d;
      min_i = i;
    }
  }
  f = fixed_point[min_i];
  p.x = floor((x - f.p.x) * pixel_size + f.x);
  p.y = floor((y - f.p.y) * pixel_size + f.y);
  return p;
}

float spot_distance(Point p1, Point p2) {
  return (float)(sqrt((p1.x - p2.x) * (p1.x - p2.x) + (p1.y - p2.y) * (p1.y - p2.y)));
}

void add_spot(int x, int y, int c) {
  Point p;
  p.x = x;
  p.y = y;
  if (c == RED_DOG) {
    if (red_count < MAX_OBJ)
      red_spot[red_count++] = p;
  } else if (c == BLUE_DOG) {
    if (blue_count < MAX_OBJ)
      blue_spot[blue_count++] = p;
  } else if (c == BALL) {
    if (ball_count < MAX_OBJ)
      ball_spot[ball_count++] = p;
  }
}

void update_mobile(Mobile *m, Point p, Point d, int angle) {
  struct timeb t;
  double ut;
  DTIME(t, ut);
  if (m->p.x == 0 && m->p.y == 0 && m->d.x == 0 && m->d.y == 0) {
    m->p.x = p.x;    m->p.y = p.y;    m->d.x = d.x;    m->d.y = d.y;    m->angle = angle;
  } else {
    m->p.x = ((m->p.x) * 0.25 + (p.x) * 0.75);
    m->p.y = ((m->p.y) * 0.25 + (p.y) * 0.75);
    m->d.x = ((m->d.x) * 0.25 + (d.x) * 0.75);
    m->d.y = ((m->d.y) * 0.25 + (d.y) * 0.75);
    m->angle = (int)((m->angle) * 0.5 + angle * 0.5);
  }
  m->t = ut;
}
