/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: vutils.c 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 *
 * Various utilities for a V4L application
 *
**/

#include "vutils.h"

/*
 * Essential structures required
 */
struct video_mmap        vid_mmap;
struct video_tuner       tuner;
struct video_capability  vcap;
struct video_buffer      vbuf;
struct video_channel     vchan;

int frame_grabber;               /* the frame buffer device */
unsigned char *frame_buffer;

int WIDTH  = 0;
int HEIGHT = 0;
int VID_BUF_SIZE = 0;
int which_frame = 1;        /* 1 implies initial sync for frame 0 */

/* 
 *  not sure where this is properly defined, but it pretty important
 */
#define BTTV_MAX_FBUF   0x208000

int tdiff(char *tag);


/********************************************************************/
/*
 * opens a V4L framegrabber and sets it up using the passed parameters
 *            device - /dev/video0/1/2/3
 *            port = 0 tv  1 composite 2 S-video  3 = composite 3
 *            norm = 0 PAL 1 NTSC  2 SECAM
 *            W, H = width & height (best to be a 4x3 ratio)
 * then sleep(1) to let the hardware settle
**/
int open_video(char *device, int port, int norm, int W, int H)
{

    /*
     *  Open the device, exit it it fails
     */
    frame_grabber = open(device, O_RDWR);
    if(frame_grabber <= 0)
    {
        perror("opening video device failed");
        return 0;
    }

    /*
     *  Query the device for its capabilities
     */
    if(ioctl(frame_grabber,VIDIOCGCAP, &vcap))
    {
        perror("VIDIOCGCAP failed ");
        close(frame_grabber);
        return 0;
    }

    /*
     *  Set capabilities, selecting correct port
     */
    vchan.channel = port;
    if(ioctl(frame_grabber, VIDIOCSCHAN, &vchan))
    {
        perror(" VIDIOCGCHAN failed");
        close(frame_grabber);
        return 0;
    }


    /*
     *  Setup the tuner
     *  which generally means the image format, NTSC or PAL
     * some cards don't have a tuner, so maybe safe to ignore an error here ??
     */
    if(ioctl(frame_grabber, VIDIOCGTUNER, &tuner)<0)
    {
        perror("VIDIOCGTUNER failed");
    }
    else
    {
        tuner.mode = norm;
        if(ioctl(frame_grabber, VIDIOCSTUNER, &tuner)<0)
        {
            perror("VIDIOCSTUNER failed");
            close(frame_grabber);
            return 0;
        }
    }

    /*
     *  Create the capture buffer for this device now we have it open
     * and basically setup
     */
    WIDTH  = W;
    HEIGHT = H;
    VID_BUF_SIZE = BTTV_MAX_FBUF; //WIDTH*HEIGHT*3; //BTTV_MAX_FBUF;
    frame_buffer = mmap(0, VID_BUF_SIZE*2,
                  PROT_READ|PROT_WRITE,
                  MAP_SHARED, frame_grabber, 0);

    if( frame_buffer == MAP_FAILED)
    {
        perror("mmap failed ");
        close(frame_grabber);
        return 0;
    }

    /*
     * setup the video_mmap struct so that we can capture sensibly
     */
    vid_mmap.frame  = 0;
    vid_mmap.height = HEIGHT;
    vid_mmap.width  = WIDTH;
    vid_mmap.format = VIDEO_PALETTE_RGB24;


    /*
     * now sleep for a time to let everything settle
     */

    sleep(1);

    vid_mmap.frame  = 0;
    if(ioctl(frame_grabber, VIDIOCMCAPTURE, &vid_mmap)<0)
    {
        perror("initial VIDIOCMCAPTURE failed");
        return 0;
    }
//    ioctl(frame_grabber,VIDIOCSYNC,&vid_mmap.frame);   /* block waiting for the frame */
       
    return frame_grabber;
}

/*********************************************************************/
/*
 * enables capturing, then get _1_ frame into the buffer and returns
 */
unsigned char *get_frame(int fd)
{

    if( which_frame == 1)  which_frame = 0;
    else                   which_frame = 1;
        
    /*  which frame will be 0 the fist time we get to this comment */
    
    if(which_frame == 1)
    {
        vid_mmap.frame  = 0;    /* ask for 0 */
        if(ioctl(frame_grabber, VIDIOCMCAPTURE, &vid_mmap)<0)
            return NULL;
        
        vid_mmap.frame  = 1;    /* sync for Frame 1 */
        ioctl(fd,VIDIOCSYNC,&vid_mmap.frame);   /* block waiting for the frame */        /* enable capturing for Frame 0 */
        return (frame_buffer+VID_BUF_SIZE);
    }   
    else
    {  
        vid_mmap.frame  = 1;    /* ask for Frame 1 */
        if(ioctl(frame_grabber, VIDIOCMCAPTURE, &vid_mmap)<0)
            return NULL;
        
        vid_mmap.frame  = 0;    /* sync for Frame 0 */
        ioctl(fd,VIDIOCSYNC,&vid_mmap.frame);   /* block waiting for the frame */
        return (frame_buffer);
    }


    return NULL;        /* we can never get here */
}

/*
 * closes the video device and un-mmaps the  buffer
 */
int close_video(int fd)
{
    munmap(frame_buffer,VID_BUF_SIZE*2 );
    close(fd);
    return TRUE;

}

/*********************************************************************/
/*
 * writes a JPEG file from a buffer to a file
 */
void write_jpeg (char *name, int width, int height, int quality, char *image)
{
int y, x, line_width;
JSAMPROW row_ptr[1];
struct jpeg_compress_struct cjpeg;
struct jpeg_error_mgr jerr;
unsigned char *line;
FILE *picture;

    picture = fopen(name, "w");
    if( picture == NULL)
        return;

    line = malloc (width * 3);
    if (!line)
        return;

    cjpeg.err = jpeg_std_error(&jerr);
    jpeg_create_compress (&cjpeg);
    cjpeg.image_width = width;
    cjpeg.image_height= height;
    cjpeg.input_components = 3;
    cjpeg.in_color_space = JCS_RGB;
    jpeg_set_defaults (&cjpeg);

    jpeg_set_quality (&cjpeg, quality, TRUE);
    cjpeg.dct_method = JDCT_FASTEST;
    jpeg_stdio_dest (&cjpeg, picture);

    jpeg_start_compress (&cjpeg, TRUE);

    row_ptr[0] = line;
    line_width = width * 3;
    for ( y = 0; y < height; y++)
    {
        for (x = 0; x < line_width; x+=3)
        {
            line[x]   = image[x+2];
            line[x+1] = image[x+1];
            line[x+2] = image[x];
        }
        jpeg_write_scanlines (&cjpeg, row_ptr, 1);
        image += line_width;
    }
    jpeg_finish_compress (&cjpeg);
    jpeg_destroy_compress (&cjpeg);
    free (line);

    fclose(picture);

    return;
}

/*********************************************************************/
/*
 * writes a colour PGM file from a buffer
 */
int write_P6(char *name, int W, int H, unsigned char *buf)
{
FILE *f;
int i, j;

    if( (f = fopen(name, "w")) == NULL)
    {
        printf("Could not open >%s<\n", name);
        return FALSE;
    }

    fprintf(f,"P6\n %d %d\n 255\n",W, H);
    fflush(f);

    for(i=0;i<H;i++) {
        for(j=0;j<W;j++)
        {
            fwrite(buf + (i * W + j)*3+2, 1, 1, f);
            fwrite(buf + (i * W + j)*3+1, 1, 1, f);
            fwrite(buf + (i * W + j)*3  , 1, 1, f);
        }
    }

    fclose(f);

    return TRUE;

}

/*
 * writes a B&W PGM file from a buffer
 */
int write_P5(char *name, int W, int H, unsigned char *buf)
{
FILE *f;
int i, j;


    if( (f = fopen(name, "w")) == NULL)
    {
        printf("Could not open >%s<\n", name);
        return FALSE;
    }

    fprintf(f,"P5\n %d %d\n 255\n",W, H);
    fflush(f);

    for(i=0;i<H;i++) {
        for(j=0;j<W;j++)
        {
              fwrite(buf + (i * W + j)*3, 1, 1, f);
        }
    }
    fclose(f);

    return TRUE;
}

