/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: ActuatorControl.cc 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
**/

#define OPENR_STUBGEN

#include <OPENR/core_macro.h>

#include <MCOOP.h>
#include <OPENR/OPENR.h>
#include <OPENR/OPENRMessages.h>
#include <OPENR/core_macro.h>

#include "ActuatorControl.h"

#include "ActionPro.h"

/*
#ifdef DELIBERATE
#include "../prolog/PrologData.h"
#endif
*/

ActuatorControl::ActuatorControl() {
  OSYSDEBUG(("ActuatorControl::ActuatorControl()\n"));
  powerIsOn = false;
  wait4settle = 30;

}

OStatus
ActuatorControl::DoInit (const OSystemEvent &)
{
  OSYSDEBUG(("ActuatorControl::DoInit()\n"));

  // NewComData  /* generates the necessary number of subjects and observers */
  // SetComData  /* registers the communication service with other objects */
  NEW_ALL_SUBJECT_AND_OBSERVER;
  REGISTER_ALL_ENTRY;
  SET_ALL_READY_AND_NOTIFY_ENTRY;


  sEffector = subject[sbjEffectors];
  sHead = subject[sbjHead];
  //sIndicators = subject[sbjIndicators];
  eCommander.init(subject[sbjECommander]);

  indicatorCommander = new Indicators((subject[sbjIndicators]), eCommander);

  initMem();
 
   // SetPAgent (subject[sbjShmPWalk]);
  return oSUCCESS;
}

OStatus
ActuatorControl::DoStart (const OSystemEvent &)
{
  OSYSDEBUG(("ActuatorControl::DoStart()\n"));

  ENABLE_ALL_SUBJECT;
  ASSERT_READY_TO_ALL_OBSERVER;
  
 /*
  for (int i = 0; i < numOfObserver; ++i)
      if (observer[i]->AssertReady () != oSUCCESS)
	  OSYSDEBUG(("AssertReady() failed. observer[%d]\n",i));

  */
  return oSUCCESS;
}

OStatus
ActuatorControl::DoStop (const OSystemEvent &)
{
  eCommander.stop();

  DISABLE_ALL_SUBJECT;
  for (int i = 0; i < numOfObserver; ++i)
    observer[i]->DeassertReady();

  return oSUCCESS;
}

OStatus
ActuatorControl::DoDestroy (const OSystemEvent &)
{
  DeleteComData         /* deletes all observers and subjects spawned
                           in DoInit() */
  return oSUCCESS;
}



void
ActuatorControl::ReadyEffectors (const OReadyEvent &)
{
	ContinueStep();
	
	//This effectively calls a GPSMotionUpdate
	subject[sbjGps]->SetData(&(pAgent->delta_forward), sizeof(pAgent->delta_forward));
	subject[sbjGps]->SetData(&(pAgent->delta_left),    sizeof(pAgent->delta_left));
	subject[sbjGps]->SetData(&(pAgent->delta_turn),    sizeof(pAgent->delta_turn));		
	subject[sbjGps]->NotifyObservers();
}

void
ActuatorControl::ReadyECommander (const OReadyEvent & event)
{
  // may not need this check; legacy comment
  if (!eCommander.isGainSet ())
    eCommander.continueInit (event);
}

void
ActuatorControl::executeAssert (const ONotifyEvent & event)
{
	execute(event);
	observer[event.ObsIndex()]->AssertReady();
}

void
ActuatorControl::setIndicationAssert (const ONotifyEvent & event)
{
	indicatorCommander->setIndication(event);
	observer[event.ObsIndex()]->AssertReady();
}

void 
ActuatorControl::ReadySetPAgent (const OReadyEvent &)
{
	SetPAgent(subject[sbjShmPWalk]);
}
	
