/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: Indicators.cc 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
**/

#include "Indicators.h"

Indicators::Indicators (OSubject * subject, EffectorCommander & eCommander) {

	sIndicators = subject;
    for (int i = 0; i < numINDICATORS; i++) {
        inds[i] = indPOS1;
    }

//    jointGainNotSet = true; Andres - new binary set

    SetUpBuffer (eCommander);
}

Indicators::Indicators () {

	//sIndicators = subject;
    for (int i = 0; i < numINDICATORS; i++) {
        inds[i] = indPOS1;
    }

//    jointGainNotSet = true; Andres - new binary set

    //SetUpBuffer (eCommander);
}


//void Indicators::setIndication (int *inds_) {
void Indicators::setIndication (const ONotifyEvent& event) {

    const int *inds_ = (const int *)event.Data(0);

/*
    static int printed = 0;
    //ERIC remove these couts later
    if (printed<0) {
		printed++;
	    cout << "this is from indicators.cc ";
    
    	for (int i = 0; i < numINDICATORS; i++) {
            cout << inds_[i] << " ";
    	}
		cout << "\n";
	}
*/

    for (int i = 0; i < numINDICATORS; i++) {
        if (inds_[i] != indUNCHANGE) {
            inds[i] = inds_[i];
        }
    }

    // Buffer still in use
    if (indRegion->NumberOfReference () != 1) {
        return;
    }

    int i = 0;

    // Mouth - JointCommand2
    for (; i < 1; i++) {
        OJointCommandValue2 *data = (OJointCommandValue2 *) indVec_->GetData (i)->value;
        for (int j = 0; j < numJOINT_IND_FRAMES; j++) {
            data[j].value = inds[i] == indPOS2 ? MOUTH_OPEN : MOUTH_CLOSED;
        }
    }

    // Tail - JointCommand2
    for (; i < 3; i++) {
        OJointCommandValue2 *data = (OJointCommandValue2 *) indVec_->GetData (i)->value;
        for (int j = 0; j < numEAR_IND_FRAMES; j++) {
	        data[j].value = inds[i];	// or UP, same value
        }
    }
    /*
    for (; i < 3; i++) {
        OJointCommandValue2 *data = (OJointCommandValue2 *) indVec_->GetData (i)->value;
        for (int j = 0; j < numEAR_IND_FRAMES; j++) {
	        switch (inds[i]) {
	        case indPOS2:
	            data[j].value = TAIL_RIGHT;	// or UP, same value
	            break;
	        case indPOS3:
	            data[j].value = TAIL_LEFT;	// or DOWN, same value
	            break;
	        default:
	            data[j].value = TAIL_CENTRED;
            }
        }
    }
    */

    // Ears - JointCommand3
    for (; i < 5; i++) {
        OJointCommandValue3 *data = (OJointCommandValue3 *) indVec_->GetData (i)->value;
        for (int j = 0; j < 5; j++) {
            data[j].value = inds[i] == indPOS1 ? ojoint3_STATE1 : ojoint3_STATE0;
        }
    }

    // LEDCommand2
    for (; i < numINDICATORS; i++) {
        OLEDCommandValue2 *data = (OLEDCommandValue2 *) indVec_->GetData (i)->value;
        data[0].led = inds[i] == indPOS1 ? oledOFF : oledON;
        data[0].period = 1;
    }
	
    sIndicators->SetData (indRegion);
    sIndicators->NotifyObservers ();
}


void Indicators::SetUpBuffer (EffectorCommander & eCommander) {

    OPENR::NewCommandVectorData (numINDICATORS, &indVecID_, &indVec_);
    indVec_->SetNumData (numINDICATORS);
    indRegion = new RCRegion (indVec_->vectorInfo.memRegionID,
			                  indVec_->vectorInfo.offset,
			                  (void *) indVec_, indVec_->vectorInfo.totalSize);

    // Sets up indicator primitives
    OPrimitiveID *indPrimID_[numINDICATORS];
    OPrimitiveID ledPrimitives[numINDICATOR_LEDS];
    for (int i = 0; i < numINDICATOR_JOINTS; i++) {
        indPrimID_[i] = eCommander.getPrimID (indicatorJoints[i]);
    }

    // Opens the led primitives and then makes the indPrimID array points to them
    for (int i = numINDICATOR_JOINTS, j = 0; i < numINDICATORS; i++, j++) {
        OPENR::OpenPrimitive ((const char *) INDICATORS[i], &ledPrimitives[j]);
        indPrimID_[i] = &ledPrimitives[j];
//cout << i << ": " << indPrimID_[i] << endl;
    }

//cout << "IndPrimID" << endl;

    int ind_i = 0;

    // Mouth and Tail - JointCommand2
    for (; ind_i < 3; ind_i++) {
        OCommandInfo *info = indVec_->GetInfo (ind_i);
        info->Set (odataJOINT_COMMAND2, *indPrimID_[ind_i],
        numJOINT_IND_FRAMES);
    }

    // Ears - JointCommand3
    for (; ind_i < 5; ind_i++) {
        OCommandInfo *info = indVec_->GetInfo (ind_i);
        info->Set (odataJOINT_COMMAND3, *indPrimID_[ind_i], numEAR_IND_FRAMES);
    }

    // LEDCommand2
    for (; ind_i < numINDICATORS; ind_i++) {
        OCommandInfo *info = indVec_->GetInfo (ind_i);
        info->Set (odataLED_COMMAND2, *indPrimID_[ind_i], 1);
    }
}

