/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: PWalk.cc 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
**/

#include <math.h>
#include <stdio.h>
#include "PWalk.h"
#include "../share/CanterCalib.h"

/*
 * PWalk
 */
PWalk::PWalk (OSubject * se, EffectorCommander & eCommander /*, int *addrForLeadingLeg, bool *addrForStepComplete*/ ) {

    cout << "This is PWalk constructor" << endl;

    sEffector = se;

    fc = 0;
    sc = 0;
    tc = 0; /* no movement */
    
    /* see thesis for parameter definitions */
    PG    = 40; 
    step_ = 2 * PG;
    hf    = 70;
    hb    = 110;
    hdf   = 5;
    hdb   = 25;
    hfG   = hf;
    hbG   = hb;
    ffo   = 55;
    fso   = 15;
    bfo   = -55;
    bso   = 15;

    theta5 = atan (119.4 / lsh);
    sin5   = sin (theta5);
    cos5   = cos (theta5);

    jfl1 = 0;
    jfl2 = 0;
    jfl3 = 0;
    jfr1 = 0;
    jfr2 = 0;
    jfr3 = 0;
    jbl1 = 0;
    jbl2 = 0;
    jbl3 = 0;
    jbr1 = 0;
    jbr2 = 0;
    jbr3 = 0;
    
    walktype = CanterWalkWT;
    
    delta_forward = delta_left = delta_turn = 0;
    
    SetUpBuffer (eCommander);
    pwalkNotUsingHead = true;
    
    //leading_leg  = addrForLeadingLeg;
    //stepComplete = addrForStepComplete;

    //*stepComplete = true;
    //*leading_leg = 0; 
    // 1 = left, -1 = right, 0 = unknown
    
    chestpush = new SpecialAction();
    chestpush->load("/MS/cpush.pos");
    fwdkick = new SpecialAction();
    fwdkick->load("/MS/fwdkick.pos");
    fastkick = new SpecialAction();
    fastkick->load("/MS/fastkick.pos");
    
    getupf = new SpecialAction();
    getupf->load("/MS/getupf.pos");
    getups = new SpecialAction();
    getups->load("/MS/getups.pos");
    getupr = new SpecialAction();
    getupr->load("/MS/getupr.pos");
    backKick = new SpecialAction();
    backKick->load("/MS/bckkick.pos");

	cheer = new SpecialAction();
	cheer->load("/MS/cheer.pos");
	
	ptr_walktype = &walktype;
	ptr_step_ = &step_;
	ptr_PG = &PG;

    readLocusOffets("/MS/offsets.txt");
}

void PWalk::pStep(int wT, double forward, double left, double turnCCW,
             int speed, double hF, double hB, double hdF, double hdB,
             double ffO, double fsO, double bfO, double bsO) {

    walktype = wT;

    for (int bufID = 0; bufID < numOfBuffer; bufID++) {
        int refcount = cmdRegion[bufID]->NumberOfReference ();

        if (refcount != 1) {
            continue;
        }

        OJointCommandValue2 *commandValue_[numOfJoint];
        for (int i = 0; i < numOfJoint; i++) {
            OCommandData *data = cmdVec_[bufID]->GetData (i);
            commandValue_[i] = (OJointCommandValue2 *) data->value;
        }

        /* Walks */
        if (walktype == NormalWalkWT || walktype == CanterWalkWT || walktype == ZoidalWalkWT || walktype == OffsetWalkWT) {
            /* Set head command to 0 frames in buffer */
            for (int i = 0; i < 3; i++) {
                OCommandInfo *info = cmdVec_[bufID]->GetInfo (i);
                info->numFrames = 0;
            }
            pwalkNotUsingHead = true;

            /* Setting Data - Legs */
            if (step_ == PG || step_ == 2 * PG) {
                /* reset step_ at end of full stepcycle */
                if (step_ >= 2 * PG) {
                    step_ = 0;
                    PG = speed;
                }

                hfG = hF;
                hbG = hB;
                ffo = ffO;
                fso = fsO;
                bfo = bfO;
                bso = bsO;
                hdf = hdF;
                hdb = hdB;

                if (walktype == CanterWalkWT) {
                    /* Calibration of canter walk */
                    
					if (left > 7.2) {
						left = 7.2;
                        cout << "Trimming left to 7.2, orig = " << left << endl;
					} else if (left < -7.2) {
						left = -7.2;
                        cout << "Trimming left to -7.2, orig = " << left << endl;
					}
					
					fc = Canter::forwardCalib(forward, left, turnCCW);
					sc = Canter::leftCalib(forward, left, turnCCW);
					tc = Canter::turnCCWCalib(forward, left, turnCCW);
                } else if (walktype == NormalWalkWT) {
                    /* Calibration of normal walk */
                    fc = (forward + .04) * 6.0;
                    sc = left * 6.0;
                    tc = turnCCW * 1.5 + sc / 6.0;
                    /* last term is to correct turning on sideways movements BH 3/8/00 */
                } else if (walktype == ZoidalWalkWT) {
                    /* Calibration of trapezoidal walk */
                    fc = (forward + .04) * 5.4;
                    sc = left * 4.5;
                    tc = turnCCW + left * 6 / 4;
                } else if (walktype == OffsetWalkWT) {
                    /* Calibration of OffsetWalkWT
                     *
                     * NOTE: this calibration is NOT ACCURATE - it has specifically been set for easiest behaviour level
                     * integration (for replacing ZoidalWalk)
                     * In particular, walking sideways is awful (most likely caused by the offsets being calibrated for
                     * walking forwards)
                     */
                    turnCCW *= 1.2;
                    fc = (forward / 8.6666*6 + .04) * 6.0;
                    sc = left * 6.0;
                    tc = turnCCW / 18.75 * 10 * 1.5 + sc / 6.0 + fc/100;
                }

                delta_forward = (double) numOfFrames / PG * forward;
                delta_left = (double) numOfFrames / PG * left;
                delta_turn = (double) numOfFrames / PG * turnCCW;
                /*
if (abs(delta_left) > 1.8 || abs(delta_forward) > 1.8 || abs(delta_turn) > 9.0)
{
  cout << "pWalk: dForward = " << delta_forward << " dLeft = " << delta_left << " dTurn = " << delta_turn << " PG=" << PG << " forward=" << forward << " left=" << left << " turnCCW=" << turnCCW << " numOfFrames=" << numOfFrames << endl;
}
*/

                /* end (step_ == PG || step_ == 2 * PG) */
            }

            /* Now calculate transform to theta 1,2 & 3
             * (see thesis for explanation)
             */
            double theta1, theta2, theta3, tempu, tempv, x, y, z, temp1, temp2;
            double theta4, f = 0, s = 0, dd = 0;
            double h = 0, fstart, fend, sstart, send, fstep, sstep, hstep;
            double PGH, PGR;	// Will: XXX: This was defined as int.  Is this change to double ok?

            /* step = step_ 180 deg out of phase */
            int step; 

            /* this is a magic part of PWalk -- canter,
             * it moves the robots' motor up and down in a sine fashion
             * such that it will moves faster in result.
             * (not used in NormalWalk and ZoidalWalk)
             */
            double stepCycle = step_ / PG * 2.0 * pi;
            if (walktype == CanterWalkWT) {
                hf = hfG + 5.0 * cos (stepCycle);
                hb = hbG + 5.0 * cos (stepCycle);
            } else {
                hf = hfG;
                hb = hbG;
            }

            /* theta4 = angle of robot body, relative to ground */
            theta4 = asin ((hb - hf) / lsh);

            for (int frame = 0; frame < numOfFrames; ++frame) {
                /*
                 * Front Right Limb
                 */

                /* work out start and end of paw in f-s plane */
                fstart = fc + tc * sin5 + ffo;
                fend   = -fc - tc * sin5 + ffo;
                sstart = -sc - tc * cos5 + fso;
                send   = sc + tc * cos5 + fso;

                /* work out locus length */
                temp1 = fstart - fend;
                temp2 = sstart - send;
                dd = sqrt (temp1 * temp1 + temp2 * temp2);

                /* step = opposite of step_ in step cycle, use
                 * step for front right, back left, and
                 * step_ for front left, back right (trot gait)
                 */
                step = step_ + PG;
                if (step >= 2 * PG) {
                    step -= 2 * PG;
                }

                if (walktype == ZoidalWalkWT) {
                    /* work out distance per step */
                    fstep = (temp1) / PG;
                    sstep = (temp2) / PG;
                    if (dd == 0) {
                        PGH = 0;
                    } else {
                        PGH = hdf / dd * PG;
                    }

                    if (PGH == 0) {
                        hstep = 0;
                    } else {
                        hstep = hdf / PGH;
                    }

                    /* Calculate points in trapezoidal locus */
                    if (step < PG / 2 - PGH) {
                        f = (fstart + fend) / 2 - step * fstep;
                        s = (sstart + send) / 2 - step * sstep;
                        h = hf;
                    }
                    if (step < PG / 2 && step >= PG / 2 - PGH) {
                        f = (fstart + fend) / 2 - step * fstep;
                        s = (sstart + send) / 2 - step * sstep;
                        h = hf - (step - PG / 2 + PGH) * hstep;
                    }
                    if (step < 3 * PG / 2 && step >= PG / 2) {
                        f = fend + (step - PG / 2) * fstep;
                        s = send + (step - PG / 2) * sstep;
                        h = hf - hdf;
                    }
                    if (step < 3 * PG / 2 + PGH && step >= 3 * PG / 2) {
                        f = fstart - (step - 3 * PG / 2) * fstep;
                        s = sstart - (step - 3 * PG / 2) * sstep;
                        h = hf - hdf + (step - 3 * PG / 2) * hstep;
                    }
                    if (step < 2 * PG && step >= 3 * PG / 2 + PGH) {
                        f = fstart - (step - 3 * PG / 2) * fstep;
                        s = sstart - (step - 3 * PG / 2) * sstep;
                        h = hf;
                    }
                } else if (walktype == NormalWalkWT || walktype == CanterWalkWT) {
                    PGR = dd == 0 ? PG : (int) ((double) PG / (1 + hdf / dd));
                    PGH = (PG - PGR) / 2;

                    /* Calculate points in rectangular locus */
                    if (step < PG / 2) {
                        fstep = (fstart - fend) / PG;
                        sstep = (sstart - send) / PG;
                        f = (fstart + fend) / 2 - step * fstep;
                        s = (sstart + send) / 2 - step * sstep;
                        h = hf;
                    }
                    if (step < PG / 2 + PGH && step >= PG / 2 && PGH != 0) {
                        f = fend;
                        s = send;
                        h = hf - (step - PG / 2) * hdf / PGH;
                    }
                    if (step < PG / 2 + PGH + PGR && step >= PG / 2 + PGH) {
                        fstep = (fstart - fend) / PGR;
                        sstep = (sstart - send) / PGR;
                        f = fend + (step - PG / 2 - PGH) * fstep;
                        s = send + (step - PG / 2 - PGH) * sstep;
                        h = hf - hdf;
                    }
                    if (step < PG / 2 + 2 * PGH + PGR
                        && step >= PG / 2 + PGH + PGR && PGH != 0) {
                        f = fstart;
                        s = sstart;
                        h = hf - hdf + hdf / PGH * (step - PG / 2 - PGH - PGR);
                    }
                    if (step < 2 * PG && step >= PG / 2 + 2 * PGH + PGR) {
                        fstep = (fstart - fend) / PG;
                        sstep = (sstart - send) / PG;
                        f = fstart - fstep * (step - PG / 2 - 2 * PGH - PGR);
                        s = sstart - sstep * (step - PG / 2 - 2 * PGH - PGR);
                        h = hf;
                    }
                } else if (walktype == OffsetWalkWT) {
                    double af, as, ah, bf, bs, bh, cf, cs, ch, df, ds, dh, ef, es, eh;
                    double d1, d2, d3, d4, d5, dtotal;
                    double t1, t2, t3, t4, t5;

                    if (dd != 0)
                    {
                        double locusRotate = atan2(fstart-fend,sstart-send) - (pi/2.0);
                        bf = fend + bsl1*sin(locusRotate) + bfl1*cos(locusRotate);
                        bs = send + bsl1*cos(locusRotate) - bfl1*sin(locusRotate);
                        cf = fend + csl1*sin(locusRotate) + cfl1*cos(locusRotate);
                        cs = send + csl1*cos(locusRotate) - cfl1*sin(locusRotate);
                        df = fstart + dsl1*sin(locusRotate) + dfl1*cos(locusRotate);
                        ds = sstart + dsl1*cos(locusRotate) - dfl1*sin(locusRotate);
                        ef = fstart + esl1*sin(locusRotate) + efl1*cos(locusRotate);
                        es = sstart + esl1*cos(locusRotate) - efl1*sin(locusRotate);
                        af = (ef + bf)/2.0;
                        as = (es + bs)/2.0;
                    }
                    else
                    {
                        bf = cf = df = ef = af = (fstart + fend)/2.0;
                        bs = cs = ds = es = as = (sstart + send)/2.0;
                    }

                    if (hdf != 0)
                    {
                        bh = hf + bhl1;
                        ch = hf + chl1 - hdf;
                        dh = hf + dhl1 - hdf;
                        eh = hf + ehl1;
                        ah = (eh + bh)/2.0;
                    }
                    else
                    {
                        bh = ch = dh = eh = ah = hf;
                    }

                    d1 = sqrt((af-bf)*(af-bf) + (as-bs)*(as-bs) + (ah-bh)*(ah-bh));
                    d2 = sqrt((bf-cf)*(bf-cf) + (bs-cs)*(bs-cs) + (bh-ch)*(bh-ch));
                    d3 = sqrt((cf-df)*(cf-df) + (cs-ds)*(cs-ds) + (ch-dh)*(ch-dh));
                    d4 = sqrt((df-ef)*(df-ef) + (ds-es)*(ds-es) + (dh-eh)*(dh-eh));
                    d5 = sqrt((ef-af)*(ef-af) + (es-as)*(es-as) + (eh-ah)*(eh-ah));
                    dtotal = d1 + d2 + d3 + d4 + d5;

                    if (dtotal != 0)
                    {
                        t1 = d1/dtotal * 2 * PG;
                        t2 = d2/dtotal * 2 * PG;
                        t3 = d3/dtotal * 2 * PG;
                        t4 = d4/dtotal * 2 * PG;
                        t5 = d5/dtotal * 2 * PG;
                    }
                    else
                    {
                        t1 = t2 = t3 = t4 = 0;
                        t5 = 2 * PG;
                    }

                    if (step < t1) {
                        fstep = (af - bf)/t1;
                        sstep = (as - bs)/t1;
                        hstep = (ah - ah)/t1;
                        f     = af - step*fstep;
                        s     = as - step*sstep;
                        h     = ah - step*hstep;
                    } else if (step < t1 + t2) {
                        fstep = (bf - cf)/t2;
                        sstep = (bs - cs)/t2;
                        hstep = (bh - ch)/t2;
                        f     = bf - (step-t1)*fstep;
                        s     = bs - (step-t1)*sstep;
                        h     = bh - (step-t1)*hstep;
                    } else if (step < t1 + t2 + t3) {
                        fstep = (cf - df)/t3;
                        sstep = (cs - ds)/t3;
                        hstep = (ch - dh)/t3;
                        f     = cf - (step-t1-t2)*fstep;
                        s     = cs - (step-t1-t2)*sstep;
                        h     = ch - (step-t1-t2)*hstep;
                    } else if (step < t1 + t2 + t3 + t4) {
                        fstep = (df - ef)/t4;
                        sstep = (ds - es)/t4;
                        hstep = (dh - eh)/t4;
                        f     = df - (step-t1-t2-t3)*fstep;
                        s     = ds - (step-t1-t2-t3)*sstep;
                        h     = dh - (step-t1-t2-t3)*hstep;
                    } else {
                        fstep = (ef - af)/t5;
                        sstep = (es - as)/t5;
                        hstep = (eh - ah)/t5;
                        f     = ef - (step-t1-t2-t3-t4)*fstep;
                        s     = es - (step-t1-t2-t3-t4)*sstep;
                        h     = eh - (step-t1-t2-t3-t4)*hstep;
                    }
                }

                /* f-s to x-y-z transformation  (need f, s and h) */
                z = s;
                y = h * cos (theta4) - f * sin (theta4);
                x = h * sin (theta4) + f * cos (theta4);

                /* x-y-z to theta1,2,3 transformation */
                temp1  = l1 * l1 + 2 * l3 * l3 + l2 * l2 - x * x - y * y - z * z;
                temp2  = 2 * sqrt ((l1 * l1 + l3 * l3) * (l2 * l2 + l3 * l3));
                theta3 = 2 * pi - atan (l1 / l3) - atan (l2 / l3) - acos (temp1 / temp2);
                tempu  = 2 * l3 * sin (theta3 / 2) * sin (theta3 / 2) + l2 * sin (theta3);
                tempv  = l1 + 2 * l3 * sin (theta3 / 2) * cos (theta3 / 2) + l2 * cos (theta3);
                theta1 = asin (z / tempv);
                temp1  = y * tempv * cos (theta1) + tempu * x;
                temp2  = tempu * tempu + tempv * tempv * cos (theta1) * cos (theta1);
                theta2 = acos (temp1 / temp2);

                /* bit of fudginess, since formula gives +/- for theta2 */
                if (abs(x - tempv * cos (theta1) * sin (theta2) - tempu * cos (theta2)) > .001)
                    theta2 *= -1;

                jfr1 = theta1;
                jfr2 = theta2;
                jfr3 = theta3;

                /*
                 * Front left Limb
                 */
                /* work out start and end of paw in f-s plane */
                fstart = fc - tc * sin5 + ffo;
                fend   = -fc + tc * sin5 + ffo;
                sstart = sc + tc * cos5 + fso;
                send   = -sc - tc * cos5 + fso;

                /* work out locus length */
                temp1 = fstart - fend;
                temp2 = sstart - send;
                dd = sqrt (temp1 * temp1 + temp2 * temp2);

                if (walktype == ZoidalWalkWT) {
                    fstep = (temp1) / PG;
                    sstep = (temp2) / PG;
                    if (dd == 0) {
                        PGH = 0;
                    } else {
                        PGH = hdf / dd * PG;
                    }

                    if (PGH == 0) {
                        hstep = 0;
                    } else {
                        hstep = hdf / PGH;
                    }

                    /* work out points on trapezoidal locus */
                    if (step_ < PG / 2 - PGH) {
                        f = (fstart + fend) / 2 - step_ * fstep;
                        s = (sstart + send) / 2 - step_ * sstep;
                        h = hf;
                    }
                    if (step_ < PG / 2 && step_ >= PG / 2 - PGH) {
                        f = (fstart + fend) / 2 - step_ * fstep;
                        s = (sstart + send) / 2 - step_ * sstep;
                        h = hf - (step_ - PG / 2 + PGH) * hstep;
                    }
                    if (step_ < 3 * PG / 2 && step_ >= PG / 2) {
                        f = fend + (step_ - PG / 2) * fstep;
                        s = send + (step_ - PG / 2) * sstep;
                        h = hf - hdf;
                    }
                    if (step_ < 3 * PG / 2 + PGH && step_ >= 3 * PG / 2) {
                        f = fstart - (step_ - 3 * PG / 2) * fstep;
                        s = sstart - (step_ - 3 * PG / 2) * sstep;
                        h = hf - hdf + (step_ - 3 * PG / 2) * hstep;
                    }
                    if (step_ < 2 * PG && step_ >= 3 * PG / 2 + PGH) {
                        f = fstart - (step_ - 3 * PG / 2) * fstep;
                        s = sstart - (step_ - 3 * PG / 2) * sstep;
                        h = hf;
                    }
                } else if (walktype == NormalWalkWT || walktype == CanterWalkWT) {
                    PGR = dd == 0 ? PG : (int) ((double) PG / (1 + hdf / dd));
                    PGH = (PG - PGR) / 2;

                    /* work out points on rectangular locus */
                    if (step_ < PG / 2) {
                        fstep = (fstart - fend) / PG;
                        sstep = (sstart - send) / PG;
                        f = (fstart + fend) / 2 - step_ * fstep;
                        s = (sstart + send) / 2 - step_ * sstep;
                        h = hf;
                    }
                    if (step_ < PG / 2 + PGH && step_ >= PG / 2 && PGH != 0) {
                        f = fend;
                        s = send;
                        h = hf - (step_ - PG / 2) * hdf / PGH;
                    }
                    if (step_ < PG / 2 + PGH + PGR && step_ >= PG / 2 + PGH) {
                        fstep = (fstart - fend) / PGR;
                        sstep = (sstart - send) / PGR;
                        f = fend + (step_ - PG / 2 - PGH) * fstep;
                        s = send + (step_ - PG / 2 - PGH) * sstep;
                        h = hf - hdf;
                    }
                    if (step_ < PG / 2 + 2 * PGH + PGR
                        && step_ >= PG / 2 + PGH + PGR && PGH != 0) {
                        f = fstart;
                        s = sstart;
                        h = hf - hdf + hdf / PGH * (step_ - PG / 2 - PGH - PGR);
                    }
                    if (step_ < 2 * PG && step_ >= PG / 2 + 2 * PGH + PGR) {
                        fstep = (fstart - fend) / PG;
                        sstep = (sstart - send) / PG;
                        f = fstart - fstep * (step_ - PG / 2 - 2 * PGH - PGR);
                        s = sstart - sstep * (step_ - PG / 2 - 2 * PGH - PGR);
                        h = hf;
                    }
                } else if (walktype == OffsetWalkWT) {
                    double af, as, ah, bf, bs, bh, cf, cs, ch, df, ds, dh, ef, es, eh;
                    double d1, d2, d3, d4, d5, dtotal;
                    double t1, t2, t3, t4, t5;

                    if (dd != 0)
                    {
                        double locusRotate = atan2(fstart-fend,sstart-send) - (pi/2.0);
                        bf = fend + bsl1*sin(locusRotate) + bfl1*cos(locusRotate);
                        bs = send + bsl1*cos(locusRotate) - bfl1*sin(locusRotate);
                        cf = fend + csl1*sin(locusRotate) + cfl1*cos(locusRotate);
                        cs = send + csl1*cos(locusRotate) - cfl1*sin(locusRotate);
                        df = fstart + dsl1*sin(locusRotate) + dfl1*cos(locusRotate);
                        ds = sstart + dsl1*cos(locusRotate) - dfl1*sin(locusRotate);
                        ef = fstart + esl1*sin(locusRotate) + efl1*cos(locusRotate);
                        es = sstart + esl1*cos(locusRotate) - efl1*sin(locusRotate);
                        af = (ef + bf)/2.0;
                        as = (es + bs)/2.0;
                    }
                    else
                    {
                        bf = cf = df = ef = af = (fstart + fend)/2.0;
                        bs = cs = ds = es = as = (sstart + send)/2.0;
                    }

                    if (hdf != 0)
                    {
                        bh = hf + bhl1;
                        ch = hf + chl1 - hdf;
                        dh = hf + dhl1 - hdf;
                        eh = hf + ehl1;
                        ah = (eh + bh)/2.0;
                    }
                    else
                    {
                        bh = ch = dh = eh = ah = hf;
                    }

                    d1 = sqrt((af-bf)*(af-bf) + (as-bs)*(as-bs) + (ah-bh)*(ah-bh));
                    d2 = sqrt((bf-cf)*(bf-cf) + (bs-cs)*(bs-cs) + (bh-ch)*(bh-ch));
                    d3 = sqrt((cf-df)*(cf-df) + (cs-ds)*(cs-ds) + (ch-dh)*(ch-dh));
                    d4 = sqrt((df-ef)*(df-ef) + (ds-es)*(ds-es) + (dh-eh)*(dh-eh));
                    d5 = sqrt((ef-af)*(ef-af) + (es-as)*(es-as) + (eh-ah)*(eh-ah));
                    dtotal = d1 + d2 + d3 + d4 + d5;

                    if (dtotal != 0)
                    {
                        t1 = d1/dtotal * 2 * PG;
                        t2 = d2/dtotal * 2 * PG;
                        t3 = d3/dtotal * 2 * PG;
                        t4 = d4/dtotal * 2 * PG;
                        t5 = d5/dtotal * 2 * PG;
                    }
                    else
                    {
                        t1 = t2 = t3 = t4 = 0;
                        t5 = 2 * PG;
                    }

                    if (step_ < t1) {
                        fstep = (af - bf)/t1;
                        sstep = (as - bs)/t1;
                        hstep = (ah - ah)/t1;
                        f     = af - step_*fstep;
                        s     = as - step_*sstep;
                        h     = ah - step_*hstep;
                    } else if (step_ < t1 + t2) {
                        fstep = (bf - cf)/t2;
                        sstep = (bs - cs)/t2;
                        hstep = (bh - ch)/t2;
                        f     = bf - (step_-t1)*fstep;
                        s     = bs - (step_-t1)*sstep;
                        h     = bh - (step_-t1)*hstep;
                    } else if (step_ < t1 + t2 + t3) {
                        fstep = (cf - df)/t3;
                        sstep = (cs - ds)/t3;
                        hstep = (ch - dh)/t3;
                        f     = cf - (step_-t1-t2)*fstep;
                        s     = cs - (step_-t1-t2)*sstep;
                        h     = ch - (step-t1-t2)*hstep;
                    } else if (step_ < t1 + t2 + t3 + t4) {
                        fstep = (df - ef)/t4;
                        sstep = (ds - es)/t4;
                        hstep = (dh - eh)/t4;
                        f     = df - (step_-t1-t2-t3)*fstep;
                        s     = ds - (step_-t1-t2-t3)*sstep;
                        h     = dh - (step_-t1-t2-t3)*hstep;
                    } else {
                        fstep = (ef - af)/t5;
                        sstep = (es - as)/t5;
                        hstep = (eh - ah)/t5;
                        f     = ef - (step_-t1-t2-t3-t4)*fstep;
                        s     = es - (step_-t1-t2-t3-t4)*sstep;
                        h     = eh - (step_-t1-t2-t3-t4)*hstep;
                    }
                }


                /* f-s to x-y-z transformation  (need f, s and h) */
                z = s;
                y = h * cos (theta4) - f * sin (theta4);
                x = h * sin (theta4) + f * cos (theta4);

                /* x-y-z to theta1,2,3 transformation */
                temp1  = l1 * l1 + 2 * l3 * l3 + l2 * l2 - x * x - y * y - z * z;
                temp2  = 2 * sqrt ((l1 * l1 + l3 * l3) * (l2 * l2 + l3 * l3));
                theta3 = 2 * pi - atan (l1 / l3) - atan (l2 / l3) - acos (temp1 / temp2);
                tempu  = 2 * l3 * sin (theta3 / 2) * sin (theta3 / 2) + l2 * sin (theta3);
                tempv  = l1 + 2 * l3 * sin (theta3 / 2) * cos (theta3 / 2) + l2 * cos (theta3);
                theta1 = asin (z / tempv);
                temp1  = y * tempv * cos (theta1) + tempu * x;
                temp2  = tempu * tempu + tempv * tempv * cos (theta1) * cos (theta1);
                theta2 = acos (temp1 / temp2);

                if (abs(x - tempv * cos (theta1) * sin (theta2) - tempu * cos (theta2)) > .001) {
                    theta2 *= -1;
                }

                jfl1 = theta1;
                jfl2 = theta2;
                jfl3 = theta3;

                /*
                 * Back left Limb
                 */

                /* work out start and end of paw in f-s plane */
                fstart = fc - tc * sin5 + bfo;
                fend   = -fc + tc * sin5 + bfo;
                sstart = sc - tc * cos5 + bso;
                send   = -sc + tc * cos5 + bso;

                /* work out locus length */
                temp1 = fstart - fend;
                temp2 = sstart - send;
                dd = sqrt (temp1 * temp1 + temp2 * temp2);

                /* step = opposite of step_ in step cycle, use
                 * step for front right, back left, and
                 * step_ for front left, back right (trot gait)
                 */
                step = step_ + PG;
                if (step >= 2 * PG) {
                    step -= 2 * PG;
                }

                if (walktype == ZoidalWalkWT) {
                    /* work out distance per step */
                    fstep = (temp1) / PG;
                    sstep = (temp2) / PG;
                    if (dd == 0) {
                        PGH = 0;
                    } else {
                        PGH = hdb / dd * PG;
                    }

                    if (PGH == 0) {
                        hstep = 0;
                    } else {
                        hstep = hdb / PGH;
                    }

                    /* work out points in trapezoidal locus */
                    if (step < PG / 2 - PGH) {
                        f = (fstart + fend) / 2 - step * fstep;
                        s = (sstart + send) / 2 - step * sstep;
                        h = hb;
                    }
                    if (step < PG / 2 && step >= PG / 2 - PGH) {
                        f = (fstart + fend) / 2 - step * fstep;
                        s = (sstart + send) / 2 - step * sstep;
                        h = hb - (step - PG / 2 + PGH) * hstep;
                    }
                    if (step < 3 * PG / 2 && step >= PG / 2) {
                        f = fend + (step - PG / 2) * fstep;
                        s = send + (step - PG / 2) * sstep;
                        h = hb - hdb;
                    }
                    if (step < 3 * PG / 2 + PGH && step >= 3 * PG / 2) {
                        f = fstart - (step - 3 * PG / 2) * fstep;
                        s = sstart - (step - 3 * PG / 2) * sstep;
                        h = hb - hdb + (step - 3 * PG / 2) * hstep;
                    }
                    if (step < 2 * PG && step >= 3 * PG / 2 + PGH) {
                        f = fstart - (step - 3 * PG / 2) * fstep;
                        s = sstart - (step - 3 * PG / 2) * sstep;
                        h = hb;
                    }
                } else if (walktype == NormalWalkWT || walktype == CanterWalkWT) {
                    PGR = dd == 0 ? PG : (int) ((double) PG / (1 + hdb / dd));
                    PGH = (PG - PGR) / 2;

                    /* work out points in rectangular locus */
                    if (step < PG / 2) {
                        fstep = (fstart - fend) / PG;
                        sstep = (sstart - send) / PG;
                        f = (fstart + fend) / 2 - step * fstep;
                        s = (sstart + send) / 2 - step * sstep;
                        h = hb;
                    }
                    if (step < PG / 2 + PGH && step >= PG / 2 && PGH != 0) {

                        f = fend;
                        s = send;
                        h = hb - (step - PG / 2) * hdb / PGH;
                    }
                    if (step < PG / 2 + PGH + PGR && step >= PG / 2 + PGH) {

                        fstep = (fstart - fend) / PGR;
                        sstep = (sstart - send) / PGR;
                        f = fend + (step - PG / 2 - PGH) * fstep;
                        s = send + (step - PG / 2 - PGH) * sstep;
                        h = hb - hdb;
                    }
                    if (step < PG / 2 + 2 * PGH + PGR
                        && step >= PG / 2 + PGH + PGR && PGH != 0) {

                        f = fstart;
                        s = sstart;
                        h = hb - hdb + hdb / PGH * (step - PG / 2 - PGH - PGR);
                    }
                    if (step < 2 * PG && step >= PG / 2 + 2 * PGH + PGR) {

                        fstep = (fstart - fend) / PG;
                        sstep = (sstart - send) / PG;
                        f = fstart - fstep * (step - PG / 2 - 2 * PGH - PGR);
                        s = sstart - sstep * (step - PG / 2 - 2 * PGH - PGR);
                        h = hb;
                    }
                } else if (walktype == OffsetWalkWT) {
                    double af, as, ah, bf, bs, bh, cf, cs, ch, df, ds, dh, ef, es, eh;
                    double d1, d2, d3, d4, d5, dtotal;
                    double t1, t2, t3, t4, t5;

                    if (dd != 0)
                    {
                        double locusRotate = atan2(fstart-fend,sstart-send) - (pi/2.0);
                        bf = fend + bsl2*sin(locusRotate) + bfl2*cos(locusRotate);
                        bs = send + bsl2*cos(locusRotate) - bfl2*sin(locusRotate);
                        cf = fend + csl2*sin(locusRotate) + cfl2*cos(locusRotate);
                        cs = send + csl2*cos(locusRotate) - cfl2*sin(locusRotate);
                        df = fstart + dsl2*sin(locusRotate) + dfl2*cos(locusRotate);
                        ds = sstart + dsl2*cos(locusRotate) - dfl2*sin(locusRotate);
                        ef = fstart + esl2*sin(locusRotate) + efl2*cos(locusRotate);
                        es = sstart + esl2*cos(locusRotate) - efl2*sin(locusRotate);
                        af = (ef + bf)/2.0;
                        as = (es + bs)/2.0;
                    }
                    else
                    {
                        bf = cf = df = ef = af = (fstart + fend)/2.0;
                        bs = cs = ds = es = as = (sstart + send)/2.0;
                    }

                    if (hdb != 0)
                    {
                        bh = hb + bhl2;
                        ch = hb + chl2 - hdb;
                        dh = hb + dhl2 - hdb;
                        eh = hb + ehl2;
                        ah = (eh + bh)/2.0;
                    }
                    else
                    {
                        bh = ch = dh = eh = ah = hb;
                    }

                    d1 = sqrt((af-bf)*(af-bf) + (as-bs)*(as-bs) + (ah-bh)*(ah-bh));
                    d2 = sqrt((bf-cf)*(bf-cf) + (bs-cs)*(bs-cs) + (bh-ch)*(bh-ch));
                    d3 = sqrt((cf-df)*(cf-df) + (cs-ds)*(cs-ds) + (ch-dh)*(ch-dh));
                    d4 = sqrt((df-ef)*(df-ef) + (ds-es)*(ds-es) + (dh-eh)*(dh-eh));
                    d5 = sqrt((ef-af)*(ef-af) + (es-as)*(es-as) + (eh-ah)*(eh-ah));
                    dtotal = d1 + d2 + d3 + d4 + d5;

                    if (dtotal != 0)
                    {
                        t1 = d1/dtotal * 2 * PG;
                        t2 = d2/dtotal * 2 * PG;
                        t3 = d3/dtotal * 2 * PG;
                        t4 = d4/dtotal * 2 * PG;
                        t5 = d5/dtotal * 2 * PG;
                    }
                    else
                    {
                        t1 = t2 = t3 = t4 = 0;
                        t5 = 2 * PG;
                    }

                    if (step < t1) {
                        fstep = (af - bf)/t1;
                        sstep = (as - bs)/t1;
                        hstep = (ah - ah)/t1;
                        f     = af - step*fstep;
                        s     = as - step*sstep;
                        h     = ah - step*hstep;
                    } else if (step < t1 + t2) {
                        fstep = (bf - cf)/t2;
                        sstep = (bs - cs)/t2;
                        hstep = (bh - ch)/t2;
                        f     = bf - (step-t1)*fstep;
                        s     = bs - (step-t1)*sstep;
                        h     = bh - (step-t1)*hstep;
                    } else if (step < t1 + t2 + t3) {
                        fstep = (cf - df)/t3;
                        sstep = (cs - ds)/t3;
                        hstep = (ch - dh)/t3;
                        f     = cf - (step-t1-t2)*fstep;
                        s     = cs - (step-t1-t2)*sstep;
                        h     = ch - (step-t1-t2)*hstep;
                    } else if (step < t1 + t2 + t3 + t4) {
                        fstep = (df - ef)/t4;
                        sstep = (ds - es)/t4;
                        hstep = (dh - eh)/t4;
                        f     = df - (step-t1-t2-t3)*fstep;
                        s     = ds - (step-t1-t2-t3)*sstep;
                        h     = dh - (step-t1-t2-t3)*hstep;
                    } else {
                        fstep = (ef - af)/t5;
                        sstep = (es - as)/t5;
                        hstep = (eh - ah)/t5;
                        f     = ef - (step-t1-t2-t3-t4)*fstep;
                        s     = es - (step-t1-t2-t3-t4)*sstep;
                        h     = eh - (step-t1-t2-t3-t4)*hstep;
                    }
                }
                

                /* f-s to x-y-z transformation  (need f, s and h) */
                z = s;
                y = h * cos (theta4) - f * sin (theta4);
                x = h * sin (theta4) + f * cos (theta4);

                /* x-y-z to theta1,2,3 transformation */
                temp1  = l1 * l1 + 2 * l3 * l3 + l4 * l4 - x * x - y * y - z * z;
                temp2  = 2 * sqrt ((l1 * l1 + l3 * l3) * (l4 * l4 + l3 * l3));
                theta3 = 2 * pi - atan (l1 / l3) - atan (l4 / l3) - acos (temp1 / temp2);
                tempu  = 2 * l3 * sin (theta3 / 2) * sin (theta3 / 2) + l4 * sin (theta3);
                tempv  = l1 + 2 * l3 * sin (theta3 / 2) * cos (theta3 / 2) + l4 * cos (theta3);
                theta1 = asin (z / tempv);
                temp1  = y * tempv * cos (theta1) - tempu * x;
                temp2  = tempu * tempu + tempv * tempv * cos (theta1) * cos (theta1);
                theta2 = acos (temp1 / temp2);

                if (abs(-x - tempv * cos (theta1) * sin (theta2) - tempu * cos (theta2)) > .001)
                    theta2 *= -1;

                jbl1 = theta1;
                jbl2 = theta2;
                jbl3 = theta3;

                /*
                 * Back Right Limb
                 */

                /* work out start and end of paw in f-s plane */
                fstart = fc + tc * sin5 + bfo;
                fend   = -fc - tc * sin5 + bfo;
                sstart = -sc + tc * cos5 + bso;
                send   = sc - tc * cos5 + bso;

                /* work out locus length */
                temp1 = fstart - fend;
                temp2 = sstart - send;
                dd = sqrt (temp1 * temp1 + temp2 * temp2);

                if (walktype == ZoidalWalkWT)
                {
                    /* work out distance per step */
                    fstep = (temp1) / PG;
                    sstep = (temp2) / PG;
                    if (dd == 0)
                        PGH = 0;
                    else
                        PGH = hdb / dd * PG;

                    if (PGH == 0)
                        hstep = 0;
                    else
                        hstep = hdb / PGH;

                    /* work out points in trapezoidal locus */
                    if (step_ < PG / 2 - PGH) {

                        f = (fstart + fend) / 2 - step_ * fstep;
                        s = (sstart + send) / 2 - step_ * sstep;
                        h = hb;
                    }
                    if (step_ < PG / 2 && step_ >= PG / 2 - PGH) {

                        f = (fstart + fend) / 2 - step_ * fstep;
                        s = (sstart + send) / 2 - step_ * sstep;
                        h = hb - (step_ - PG / 2 + PGH) * hstep;
                    }
                    if (step_ < 3 * PG / 2 && step_ >= PG / 2) {

                        f = fend + (step_ - PG / 2) * fstep;
                        s = send + (step_ - PG / 2) * sstep;
                        h = hb - hdb;
                    }
                    if (step_ < 3 * PG / 2 + PGH && step_ >= 3 * PG / 2) {

                        f = fstart - (step_ - 3 * PG / 2) * fstep;
                        s = sstart - (step_ - 3 * PG / 2) * sstep;
                        h = hb - hdb + (step_ - 3 * PG / 2) * hstep;
                    }
                    if (step_ < 2 * PG && step_ >= 3 * PG / 2 + PGH) {

                        f = fstart - (step_ - 3 * PG / 2) * fstep;
                        s = sstart - (step_ - 3 * PG / 2) * sstep;
                        h = hb;
                    }
                }
                else if (walktype == NormalWalkWT || walktype == CanterWalkWT)
                {
                    PGR = dd == 0 ? PG : (int) ((double) PG / (1 + hdb / dd));
                    PGH = (PG - PGR) / 2;

                    /* work out points in rectangular locus */
                    if (step_ < PG / 2) {

                        fstep = (fstart - fend) / PG;
                        sstep = (sstart - send) / PG;
                        f = (fstart + fend) / 2 - step_ * fstep;
                        s = (sstart + send) / 2 - step_ * sstep;
                        h = hb;
                    }
                    if (step_ < PG / 2 + PGH && step_ >= PG / 2 && PGH != 0) {

                        f = fend;
                        s = send;
                        h = hb - (step_ - PG / 2) * hdb / PGH;
                    }
                    if (step_ < PG / 2 + PGH + PGR && step_ >= PG / 2 + PGH) {

                        fstep = (fstart - fend) / PGR;
                        sstep = (sstart - send) / PGR;
                        f = fend + (step_ - PG / 2 - PGH) * fstep;
                        s = send + (step_ - PG / 2 - PGH) * sstep;
                        h = hb - hdb;
                    }
                    if (step_ < PG / 2 + 2 * PGH + PGR
                        && step_ >= PG / 2 + PGH + PGR && PGH != 0) {

                        f = fstart;
                        s = sstart;
                        h = hb - hdb + hdb / PGH * (step_ - PG / 2 - PGH - PGR);
                    }
                    if (step_ < 2 * PG && step_ >= PG / 2 + 2 * PGH + PGR) {

                        fstep = (fstart - fend) / PG;
                        sstep = (sstart - send) / PG;
                        f = fstart - fstep * (step_ - PG / 2 - 2 * PGH - PGR);
                        s = sstart - sstep * (step_ - PG / 2 - 2 * PGH - PGR);
                        h = hb;
                    }
                }
                else if (walktype == OffsetWalkWT)
                {
                    double af, as, ah, bf, bs, bh, cf, cs, ch, df, ds, dh, ef, es, eh;
                    double d1, d2, d3, d4, d5, dtotal;
                    double t1, t2, t3, t4, t5;

                    if (dd != 0)
                    {
                        double locusRotate = atan2(fstart-fend,sstart-send) - (pi/2.0);
                        bf = fend + bsl2*sin(locusRotate) + bfl2*cos(locusRotate);
                        bs = send + bsl2*cos(locusRotate) - bfl2*sin(locusRotate);
                        cf = fend + csl2*sin(locusRotate) + cfl2*cos(locusRotate);
                        cs = send + csl2*cos(locusRotate) - cfl2*sin(locusRotate);
                        df = fstart + dsl2*sin(locusRotate) + dfl2*cos(locusRotate);
                        ds = sstart + dsl2*cos(locusRotate) - dfl2*sin(locusRotate);
                        ef = fstart + esl2*sin(locusRotate) + efl2*cos(locusRotate);
                        es = sstart + esl2*cos(locusRotate) - efl2*sin(locusRotate);
                        af = (ef + bf)/2.0;
                        as = (es + bs)/2.0;
                    }
                    else
                    {
                        bf = cf = df = ef = af = (fstart + fend)/2.0;
                        bs = cs = ds = es = as = (sstart + send)/2.0;
                    }
                    
                    if (hdb != 0)
                    {
                        bh = hb + bhl2;
                        ch = hb + chl2 - hdb;
                        dh = hb + dhl2 - hdb;
                        eh = hb + ehl2;
                        ah = (eh + bh)/2.0;
                    }
                    else
                    {
                        bh = ch = dh = eh = ah = hb;
                    }

                    d1 = sqrt((af-bf)*(af-bf) + (as-bs)*(as-bs) + (ah-bh)*(ah-bh));
                    d2 = sqrt((bf-cf)*(bf-cf) + (bs-cs)*(bs-cs) + (bh-ch)*(bh-ch));
                    d3 = sqrt((cf-df)*(cf-df) + (cs-ds)*(cs-ds) + (ch-dh)*(ch-dh));
                    d4 = sqrt((df-ef)*(df-ef) + (ds-es)*(ds-es) + (dh-eh)*(dh-eh));
                    d5 = sqrt((ef-af)*(ef-af) + (es-as)*(es-as) + (eh-ah)*(eh-ah));
                    dtotal = d1 + d2 + d3 + d4 + d5;

                    if (dtotal != 0)
                    {
                        t1 = d1/dtotal * 2 * PG;
                        t2 = d2/dtotal * 2 * PG;
                        t3 = d3/dtotal * 2 * PG;
                        t4 = d4/dtotal * 2 * PG;
                        t5 = d5/dtotal * 2 * PG;
                    }
                    else
                    {
                        t1 = t2 = t3 = t4 = 0;
                        t5 = 2 * PG;
                    }

                    if (step_ < t1) {
                        fstep = (af - bf)/t1;
                        sstep = (as - bs)/t1;
                        hstep = (ah - ah)/t1;
                        f     = af - step_*fstep;
                        s     = as - step_*sstep;
                        h     = ah - step_*hstep;
                    } else if (step_ < t1 + t2) {
                        fstep = (bf - cf)/t2;
                        sstep = (bs - cs)/t2;
                        hstep = (bh - ch)/t2;
                        f     = bf - (step_-t1)*fstep;
                        s     = bs - (step_-t1)*sstep;
                        h     = bh - (step_-t1)*hstep;
                    } else if (step_ < t1 + t2 + t3) {
                        fstep = (cf - df)/t3;
                        sstep = (cs - ds)/t3;
                        hstep = (ch - dh)/t3;
                        f     = cf - (step_-t1-t2)*fstep;
                        s     = cs - (step_-t1-t2)*sstep;
                        h     = ch - (step_-t1-t2)*hstep;
                    } else if (step_ < t1 + t2 + t3 + t4) {
                        fstep = (df - ef)/t4;
                        sstep = (ds - es)/t4;
                        hstep = (dh - eh)/t4;
                        f     = df - (step_-t1-t2-t3)*fstep;
                        s     = ds - (step_-t1-t2-t3)*sstep;
                        h     = dh - (step_-t1-t2-t3)*hstep;
                    } else {
                        fstep = (ef - af)/t5;
                        sstep = (es - as)/t5;
                        hstep = (eh - ah)/t5;
                        f     = ef - (step_-t1-t2-t3-t4)*fstep;
                        s     = es - (step_-t1-t2-t3-t4)*sstep;
                        h     = eh - (step_-t1-t2-t3-t4)*hstep;
                    }
                }

                /* f-s to x-y-z transformation  (need f, s and h) */
                z = s;
                y = h * cos (theta4) - f * sin (theta4);
                x = h * sin (theta4) + f * cos (theta4);

                /* x-y-z to theta1,2,3 transformation */
                temp1  = l1 * l1 + 2 * l3 * l3 + l4 * l4 - x * x - y * y - z * z;
                temp2  = 2 * sqrt ((l1 * l1 + l3 * l3) * (l4 * l4 + l3 * l3));
                theta3 = 2 * pi - atan (l1 / l3) - atan (l4 / l3) - acos (temp1 / temp2);
                tempu  = 2 * l3 * sin (theta3 / 2) * sin (theta3 / 2) + l4 * sin (theta3);
                tempv  = l1 + 2 * l3 * sin (theta3 / 2) * cos (theta3 / 2) + l4 * cos (theta3);
                theta1 = asin (z / tempv);
                temp1  = y * tempv * cos (theta1) - tempu * x;
                temp2  = tempu * tempu + tempv * tempv * cos (theta1) * cos (theta1);
                theta2 = acos (temp1 / temp2);

                if (abs(-x - tempv * cos (theta1) * sin (theta2) - tempu * cos (theta2)) > .001) {
                    theta2 *= -1;
                }

                jbr1 = theta1;
                jbr2 = theta2;
                jbr3 = theta3;

                /*
                 * increment step_, fill motor buffers
                 */
                step_++;

                for (int i = 3; i < numOfJoint; ++i) {
                    switch (i) {

                        case 3:  commandValue_[i][frame].value = micro (jfr2); break;
                        case 4:  commandValue_[i][frame].value = micro (jfr1); break;
                        case 5:  commandValue_[i][frame].value = micro (jfr3); break;
                        case 6:  commandValue_[i][frame].value = micro (jfl2); break;
                        case 7:  commandValue_[i][frame].value = micro (jfl1); break;
                        case 8:  commandValue_[i][frame].value = micro (jfl3); break;
                        case 9:  commandValue_[i][frame].value = micro (jbr2); break;
                        case 10: commandValue_[i][frame].value = micro (jbr1); break;
                        case 11: commandValue_[i][frame].value = micro (jbr3); break;
                        case 12: commandValue_[i][frame].value = micro (jbl2); break;
                        case 13: commandValue_[i][frame].value = micro (jbl1); break;
                        case 14: commandValue_[i][frame].value = micro (jbl3); break;
                    }
                } /* end i=joint loop */
            } /* end frame loop */
        } /* end walks */

        /* Special Actions */
        else {
            SpecialAction *sp = chestpush;
            int posaIndex = 0;
            
            delta_forward = delta_left = delta_turn = 0;
            
            if (walktype == ChestPushWT) {
                sp = chestpush;
            } else if (walktype == FwdKickWT) {
                sp = fwdkick;
            } else if (walktype == FastKickWT) {
                sp = fastkick;
            } else if (walktype == GetupFrontWT) {
	            sp = getupf;
	        } else if (walktype == GetupSideWT) {
	            sp = getups;
	        } else if (walktype == GetupRearWT) {
	            sp = getupr;
	        } else if (walktype == BackKickWT) {
	            sp = backKick;
	        } else if (walktype == CheerWT) {
				sp = cheer;
			}


            for (int i = 0; i < 3; ++i) {
                OCommandInfo *info = cmdVec_[bufID]->GetInfo(i);
                info->numFrames = numOfFrames;
            }
            pwalkNotUsingHead = false;

            if (sp->step >= sp->duration) {
                sp->step = 0;
            } else {
                int cumulativeDuration = 0;
                for (int i = 0; i < sp->numOfSteps; ++i) {
                    cumulativeDuration += sp->posa[i*posaLength];
                    if (sp->step < cumulativeDuration) {
                        posaIndex = i;
                        break;
                    }
                }
            }

            for (int frame = 0; frame < numOfFrames; ++frame) {
                for (int i = 1; i < posaLength; ++i) {
                    commandValue_[i-1][frame].value = microRad( sp->posa[posaIndex*posaLength+i] );
                }
            }

            sp->step++;
        } /* end special actions */
        
        sEffector->SetData (cmdRegion[bufID]);
    }  /* end bufID loop */
    sEffector->NotifyObservers ();
}  /* (finally) end function pStep */

void PWalk::SetUpBuffer (EffectorCommander & eCommander) {

    /* Set that each data is used for a Joint control */
    for (int j = 0; j < numOfBuffer; ++j) {

        OPENR::NewCommandVectorData (numOfJoint, &cmdVecID_[j], &cmdVec_[j]);
        cmdVec_[j]->SetNumData (numOfJoint);
        cmdRegion[j] = new RCRegion (cmdVec_[j]->vectorInfo.memRegionID,
                                     cmdVec_[j]->vectorInfo.offset,
                                     (void *) cmdVec_[j],
                                     cmdVec_[j]->vectorInfo.totalSize);
    }

    /* 
     * Get joint primitve so that data can to be sent using OVRComm
     * For the definition of Joint,  refer to RL2.h.
     * Prepare a buffer where data to OVRComm is written.
     * Set so that the joint pass is used for a Joint control
     */
    OPrimitiveID *walkPrimID[numOfJoint];
    for (int i = 0; i < numOfJoint; ++i) {

        walkPrimID[i] = eCommander.getPrimID (walkJoints[i]);
        OCommandInfo *info = cmdVec_[0]->GetInfo (i);
        info->Set (odataJOINT_COMMAND2, *walkPrimID[i], numOfFrames);
        info = cmdVec_[1]->GetInfo (i);
        info->Set (odataJOINT_COMMAND2, *walkPrimID[i], numOfFrames);
    }
}

void
PWalk::readLocusOffets(const char *filename)
{
    ifstream in(filename);
    if (in)
    {
        in >> bfl1 >> bsl1 >> bhl1 >>
            cfl1 >> csl1 >> chl1 >>
            dfl1 >> dsl1 >> dhl1 >>
            efl1 >> esl1 >> ehl1 >>
            bfl2 >> bsl2 >> bhl2 >>
            cfl2 >> csl2 >> chl2 >>
            dfl2 >> dsl2 >> dhl2 >>
            efl2 >> esl2 >> ehl2;
    }
    else
    {
        /* unable to open file, default to 0 offsets (uniform
         * speed rectangular locus
         */
        cout << "Unable to read locus offsets, reverting to 0 offsets." << endl;
        bfl1 = bsl1 = bhl1 =
        cfl1 = csl1 = chl1 =
        dfl1 = dsl1 = dhl1 =
        efl1 = esl1 = ehl1 =
        bfl2 = bsl2 = bhl2 =
        cfl2 = csl2 = chl2 =
        dfl2 = dsl2 = dhl2 =
        efl2 = esl2 = ehl2 = 0;
    }
}



slongword PWalk::microRad (double radius) {
    return micro (radius / 180.0 * pi);
}

double PWalk::abs (double num) {

    if (num < 0) {
        return (-num);
    } else {
        return num;
    }
}

double PWalk::microRad2Deg (slongword rad) {
    return rad * 180.0 / (1000000.0 * pi);
}

/*
 * end PWalk
 */
 
 
