/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: Behaviours.cc 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
**/

#include "Behaviours.h"
#include "allPlayers.h"
#ifndef OFFLINE
#include "../vision/Vision.h"
#endif // OFFLINE
#include <fstream>

static const int DEFAULT_PLAYER_NUMBER = 1;

static const int REMOTE_ONLY       = 0;
static const int TEAMA_GOALIE      = 1;
static const int TEAMA_FORWARD     = 2;
static const int TEAMB_GOALIE      = 3;
static const int TEAMB_FORWARD     = 4;
static const int BALL_COLLECTION   = 5;
static const int TEST_TURNKICK     = 6;
static const int LOC_CHALLENGE     = 7;
static const int BALL_TRACK        = 8;
static const int TEAMC_GOALIE      = 9;
static const int MALAK_THE_AVENGER = 10;
static const int TEST_ORIENT       = 11;
static const int TEAMD_FORWARD     = 12;
static const int OBS_CHALLENGE     = 13;
static const int TEAME_FORWARD     = 14;
static const int MAVERICK          = 15;
static const int ZANN_THE_DEFENDER = 16;
static const int FIRE_PAWKICK      = 17;
static const int MAVERICK_GOALIE   = 18;
static const int BECKHAM_FORWARD   = 19;
static const int BECKHAM_GOALIE    = 20;
static const int PENALTY_SHOOTER   = 21;
static const int GERMAN_GOALIE     = 22;
static const int NEWOBS_CHALLENGE  = 23;
static const int ODOMETRY_TESTER   = 24;

static const int NORMAL_STANCE  = 0;
static const int CHEST_PUSH     = 1;
static const int GOALIE_KICK    = 2;
static const int LIGHTNING_KICK = 3;
static const int SPIN_KICK      = 4;
static const int TURN_KICK      = 5;
static const int BLOCK_STANCE   = 6;
static const int HOLD_STANCE    = 7;

// static variables used when TURN_KICK is set to cmdStance
static const int NO_TURN     = 0;
static const int NORMAL_TURN = 1;
static const int VSOFT_TURN  = 2;
static const int SMALL_TURN  = 3;
static const int MED_TURN    = 4;
static const int FULL_TURN   = 5;

static const bool delayDebugMsg   = false;
static const bool emonLogDebugMsg = false;

static bool tempForStepComplete = true;

// constants for locating ball while still
static const int BALL_LOST = 5;
static const int INTERVAL = 20;

// store the default behaviour on startup
#ifdef OFFLINE
static const char *DEFAULT_PLAYER_FILE = "../../robot/cfg/player.cfg";
#else // OFFLINE
static const char *DEFAULT_PLAYER_FILE = "/MS/player.cfg";
#endif // OFFLINE

// generate the assembly code of the program when it crash
#ifdef OFFLINE
static const char *EMON_LOG_FILE = "emon.log"; // shouldn't be used
#else // OFFLINE
static const char *EMON_LOG_FILE = "/MS/OPEN-R/emon.log";
#endif // OFFLINE

// a notice to indicate the clear of the emon.log
static const char *EMON_LOG_CLEARED_NOTICE = "This file is clear by Behaviours::clearEmonLog";

double Behaviours::cpuUsage = 0;
long Behaviours::currentFrameID = 0;
long Behaviours::lastFrameID = 0;
static int tail_down_count = 0; //for CPU time

void Behaviours::InitBehaviours() {
    clearEmonLog();

#ifdef OFFLINE
	mode_ = mNormal;
	cout << "setting mode to normal ..." << endl;
#else // OFFLINE
    mode_ = mDemobilized;
#endif // OFFLINE
    modeBeforeFallen = mDemobilized;
    playerIndicatorNotSet = true;
    actionTimer = 0;
    destPan = 85;
    frame = 0;

    PLAYER_NUMBER = DEFAULT_PLAYER_NUMBER;
    //assignMode = 0;

    // SetUpSounds();
    
    //by default (just incase we are using pstep pull interface) just make  stepComplete point
    //to something that is always true... if we are using the push interface Vision.cc with reset
    //stepComplete to point to something more sensible

    stepComplete = &tempForStepComplete;

    loadDefaultPlayer(DEFAULT_PLAYER_FILE);
    initIndicators();
    initPlayer();
    
    // store the old player the same as player in case the control is released the first time
    oldPLAYER = PLAYER;
    
    ballChallenge::calTurn  = -30.0;

    //these are all used to calibrate the proper variable turn kick
    pvtkDuration = 40;
    pvtkDirection = -1;
    step_ = 1;
    headLiftTime = 17;
    turnAmount = 50;
    errorMargin = 4;
    applyForwardUntilThisFromEnd = 8;
    forwardVector = 1.5;
    startStep_ = 55;
    startForwardVector = 4;
}

void Behaviours::initPlayer() {
    switch (PLAYER) {
        case BALL_COLLECTION:
            ballChallenge::initBallCollection();
            break;
        case LOC_CHALLENGE:
            locChallenge::initLocChallenge();
            break;
        case BALL_TRACK:
            break;
        case TEST_ORIENT:
            break;
        case OBS_CHALLENGE:
            obstacleChallenge::initObsChallenge();
            break;
        case BECKHAM_FORWARD:    
            Beckham::initBeckhamForward();
            break;
        case BECKHAM_GOALIE:
            Beckham::initBeckhamGoalie();
            break;
        case PENALTY_SHOOTER:
            PenaltyShooter::initPenaltyShooter();
            break;
		case GERMAN_GOALIE:
	    	German::initGermanGoalie();
	    	break;
        case NEWOBS_CHALLENGE:
            newobstacleChallenge::initNewObsChallenge();
            break;
		case ODOMETRY_TESTER:
			Odometry::initOdometryTester();
			break;
        default:
            initRemoteControl();
            break;
    }

    printCurrentPlayer();
}

void Behaviours::initRemoteControl() {
#ifdef SENDING_DEBUG
#ifdef SENDING_CPLANE
    cmdCPlane             = 0;
#endif // SENDING_CPLANE
#ifdef SENDING_WORLD_MODEL_DEBUG
    cmdWDisplay           = 0;
#endif // SENDING_WORLD_MODEL_DEBUG
#ifdef SENDING_OPLANE
    cmdOPlane             = 0;
#endif // SENDING_OPLANE
#endif // SENDING_DEBUG

    cmdStationaryLocalise = 0;
    cmdActiveGpsLocalise  = 0;
    cmdStance             = NORMAL_STANCE;
    cmdTurnKick           = 0;
    resetMovement();
}

void Behaviours::Reset() {
    initPlayer();
}

void Behaviours::DecideNextAction() {
    
	vision->player = PLAYER;
	
	vision->isGoalie = false;
	/*
    if (PLAYER == BECKHAM_GOALIE)
    {
        vision->isGoalie = true;
    }
    else
    {
        vision->isGoalie = false;
    }
	*/

    switch (PLAYER) {
        case LOC_CHALLENGE:
            locChallenge::doLocChallenge();
            break;
        case BALL_COLLECTION:
            ballChallenge::doBallCollection();
            break;
        case BALL_TRACK:
            saBallTracking(true);
            setStandParams();
            break;
        case TEST_ORIENT:
            setRotateParams();
            break;
        case OBS_CHALLENGE:
            obstacleChallenge::doObsChallenge();
            break;
        case BECKHAM_FORWARD:
            Beckham::doBeckhamForward();
            break;
        case BECKHAM_GOALIE:
            Beckham::doBeckhamGoalie();
            break;
        case PENALTY_SHOOTER:
            PenaltyShooter::doPenaltyShooter();
            break;
		case GERMAN_GOALIE:
			German::doGermanGoalie();
			break;	
        case NEWOBS_CHALLENGE:
            newobstacleChallenge::doNewObsChallenge();
            break;		
		case ODOMETRY_TESTER:
			Odometry::doOdometryTester();
        	break;
		default:
            setDefaultParams();
            break;
    }

    processWControl();
    
    if (*stepComplete) {
        switch (walkType) {
        case ChestPushWT:
            makeChestPush();
            break;

        case HoldBallWT:
            makeBallHold();
            break;

        case ZoidalWalkWT:
        	// hack zoidal to actually use offsetWalk
            if (forward != 0 || left != 0 || turnCCW != 0) {
                if (hdF == 0 && hdB == 0) {
                    hdF = 5;
                    hdB = 25;
                }
                ffO = 60;
                fsO = 5;
                bsO = 10;
            }
            makeParaWalk(OffsetWalkWT, forward, left, turnCCW, PG, 
                         hF, hB, hdF, hdB, ffO, fsO, bfO, bsO);
            break;

        default:
            if (forward != 0 || left != 0 || turnCCW != 0) {
                if (hdF == 0 && hdB == 0) {
                    hdF = 10;
                    hdB = 25;
                }
                /*
                if (walkType == CanterWalkWT) {
 						cout << "flt, " << forward << ", " << left << ", " << turnCCW << endl;
                }
                */
           }

			if(standTall)
				hF = 85;
            makeParaWalk (walkType, forward, left, turnCCW, PG, 
                          hF, hB, hdF, hdB, ffO, fsO, bfO, bsO);
            break;
        }
    }
    
    
    if(gps->canSee(vobBall)) {
        ear_l = indEARUP;
        ear_r = indEARUP;
    } else {
        ear_l = indEARDOWN;
        ear_r = indEARDOWN;
    }
    
    makeHead(headtype, panx, tilty, nullz);
    
    const double CPU_OVER = (1.0 / 25.0) * 1000000.0;
    const double CPU_80 = CPU_OVER * 0.8;
    const double FRAME_SKIP = 10;
    const int TAIL_DOWN_DURATION = 25;
    static const int TAIL_INCREMENT_25 = (TAIL_RIGHT-TAIL_LEFT)/25;
    
    //Tail indicator for CPU time
    if (tail_down_count <= 0) {
        if (currentFrameID - lastFrameID >= FRAME_SKIP && lastFrameID > 0) {
            tail_v = TAIL_DOWN;
            tail_down_count = TAIL_DOWN_DURATION;
            cout << "FRAME DROPPED\n";
        } else {
            if (cpuUsage < CPU_80) {
                tail_v = TAIL_UP;
            } else {
                tail_v = MAX((TAIL_UP - (int)(((cpuUsage - CPU_80) / (CPU_OVER - CPU_80)) * TAIL_UP)), 0);
            }
        }
    } else {
        tail_v = TAIL_DOWN;
        tail_down_count--;
    }

    //Tail indicator for number of NightOwl frames dropped in the last 25 frames (right for none, left for 25)
    //tail_h = TAIL_RIGHT - (TAIL_INCREMENT_25 * (vision->lastAbortCount));
      
    lastFrameID = currentFrameID;
    
#ifndef OFFLINE
    OPowerStatus currentStatus;
    OPENR::GetPowerStatus(&currentStatus);
//    cout << "Battery current: " << currentStatus.current << "mA" << endl;
    if(currentStatus.robotStatus & orsbBATTERY_OVER_CURRENT) {
        led1 = 2;
        led2 = 2;
        led3 = 2;
        led4 = 2;
        led5 = 2;
        led6 = 2;
        barLed = 2;
    }
#endif // OFFLINE
    makeIndicators(mouth, tail_h, tail_v, ear_l, ear_r, 
                   led1, led2, led3, led4, led5, led6, barLed, tailBlue, tailRed);
    
    prevAttackMode = attackMode;

}

// this function is only needed because it is referred by GrabBall, SpinChestPush and SpinKick.
// It is not referred in any of teamA or TeamB strategies.
void Behaviours::chooseStrategy() {
    switch (PLAYER) {
        /*
        case TEAMB_GOALIE:
            TeamB::chooseTeamBGoalieStrategy();
            break;
        case TEAMC_GOALIE:
            TeamC::chooseTeamCGoalieStrategy();
            break;
        case TEAMD_FORWARD:
            TeamD::chooseTeamDForwardStrategy();
            break;
        case TEAME_FORWARD:
            TeamE::chooseTeamEForwardStrategy();
            break;
        case MALAK_THE_AVENGER:
            uther::chooseMalakTheAvengerStrategy();
            break;
        case ZANN_THE_DEFENDER:
            uther::chooseZannTheDefenderStrategy();
            break;
        */
        default:
        case BECKHAM_FORWARD:    
            Beckham::chooseBeckhamForwardStrategy();
            break;
    }
}

void Behaviours::chooseShootStrategy() {
    switch (PLAYER) {
        case BECKHAM_FORWARD:    
            Beckham::chooseBeckhamShootStrategy();
            break;
    }
}

void Behaviours::setDefaultParams() {

    // Reset Walk paramters if previous step has completed
    if (*stepComplete) {
        setStandParams();
    }
    // Head parameters
    headtype = ABS_H;
    tilty    = 0;
    panx     = 0;

    // Indicator
    mouth    = cmdMouth;
    tail_h   = cmdTailH;
    tail_v   = cmdTailV;
    ear_l    = cmdEarL;
    ear_r    = cmdEarR;
    led1     = cmdLed1;
    led2     = cmdLed2;
    led3     = cmdLed3;
    led4     = cmdLed4;
    led5     = cmdLed5;
    led6     = cmdLed6;
    barLed   = cmdBarLed;
    tailBlue = cmdTailBlue;
    tailRed  = cmdTailRed;
}

void Behaviours::headBallTracking() {
    double x1,y1,z1,x2,y2,z2,x3,y3,z3,x4,y4,z4,x5,z5;

    x1 = vision->vob[vobBall].d * tan(radians(vision->vob[vobBall].imgHead));
    y1 = vision->vob[vobBall].d * tan(radians(vision->vob[vobBall].imgElev));
    z1 = vision->vob[vobBall].d;

    x2 = x1;
    y2 = y1 + NECK_LENGTH2;
    z2 = z1 + FACE_LENGTH2;

    x3 = x2*cos(radians(-hPan))-z2*sin(radians(-hPan));
    y3 = y2;
    z3 = x2*sin(radians(-hPan))+z2*cos(radians(-hPan));

    x4 = x3;
    y4 = z3*sin(radians(hTilt))+y3*cos(radians(hTilt));
    z4 = z3*cos(radians(hTilt))-y3*sin(radians(hTilt));

    if (z4 < NECK_LENGTH) {
        z4 = NECK_LENGTH;
    }

    headtype = ABS_H;
    tilty = degrees(atan(y4/z4)-asin(NECK_LENGTH/sqrt(z4*z4 + y4*y4)));

    x5 = x4;
    
    z5 = z4*cos(radians(-tilty))-y4*sin(radians(-tilty));
    
    panx = degrees(atan(x5/z5));

    // need this bit?
    tilty = hTilt+(tilty-hTilt)*0.6;
    panx  = hPan+(panx-hPan)*0.8;
}


void Behaviours::findBallHeadParams() {
    headtype = ABS_H;
    if (panDirection == 0 && tiltDirection == 0) {
        if (gps->self().pos.x < 145.0) {
            panDirection = -1;
        } else {
            panDirection = 1;
        }
        lastTilt = -30.0;
        lastPan = panDirection * 8.0;
    } else {
        if (panDirection != 0) {
            if (lastPan <= -70 || lastPan >= 70) {
                if (lastTilt <= -30) {
                    tiltDirection = 1;
                } else if (lastTilt >= 0) {
                    tiltDirection = -1;
                }
                panDirection = 0;
            }
        } else if (tiltDirection != 0) {
            if (lastTilt <= -30 || lastTilt >= 0) {
                if (lastPan <= -70) {
                    panDirection = 1;
                } else if (lastPan >= 70) {
                    panDirection = -1;
                }
                tiltDirection = 0;
            }
        }
        lastPan  += panDirection * 8.0;
        lastTilt += tiltDirection * 8.0;
    }
    
    tilty = lastTilt;
    panx  = lastPan;

}


void Behaviours::setGotoBallParams() {
    // need a lot of turn to get align with ball
    if (ABS(vision->vob[vobBall].h) > 30) {
        forward = Cap(vision->vob[vobBall].d,2.0) * cos(DEG2RAD(vision->vob[vobBall].h));
        left    = Cap(vision->vob[vobBall].d,2.0) * sin(DEG2RAD(vision->vob[vobBall].h));
        turnCCW = Cap(vision->vob[vobBall].h, 30.0);
    } else {
        // can see the ball comfortably
        if (vision->vob[vobBall].d > 30.0) {
            forward = Cap(vision->vob[vobBall].d,8.0) * cos(DEG2RAD(vision->vob[vobBall].h));
            left    = Cap(vision->vob[vobBall].d,6.0) * sin(DEG2RAD(vision->vob[vobBall].h));
        } else {
            forward = Cap(vision->vob[vobBall].d,6.0) * cos(DEG2RAD(vision->vob[vobBall].h));
            left    = Cap(vision->vob[vobBall].d,5.0) * sin(DEG2RAD(vision->vob[vobBall].h));
        }

        double move = sqrt(SQUARE(forward) + SQUARE(left));
        double maxTurn = 9 - move;

        turnCCW = Cap(vision->vob[vobBall].h/2.0, maxTurn);
    }
}

void Behaviours::processWControl() {

    double doubleValue;
    int intValue, intValue2;
    
    doubleValue = command.value;
    intValue    = (int)(doubleValue);
    switch(command.RC_COMMAND) {
    
        case PLAYER_CHANGE:
            if (PLAYER != intValue) { // switch if player is different
                PLAYER = intValue;
                initPlayer();
            }
            break;
            
        case FORWARD_MOVE:
            cmdForward = doubleValue;
            break;
            
        case LEFT_MOVE:
            cmdLeft = doubleValue;
            break;
            
        case TURN_MOVE:
            cmdTurn = doubleValue;
            break;
            
        case SET_TILT:
            // to prevent strange behaviour when this is activated after 'u' or 'k'
            if (cmdHeadtype==ABS_H || cmdHeadtype==ABS_PT) {
                cmdPan = 0;
            }
            cmdTilt     = doubleValue;
            cmdHeadtype = REL_H;
            break;
            
        case SET_PAN:
            // to prevent strange behaviour when this is activated after 'u' or 'k'
            if (cmdHeadtype==ABS_H || cmdHeadtype==ABS_PT) {
                cmdTilt = 0;
            }
            cmdPan     = doubleValue;
            cmdHeadtype = REL_H;
            break;
            
        case DO_BLOCK_STANCE:
            cmdStance = BLOCK_STANCE;
            resetMovement();
            cmdHeadtype = REL_H;
            break;
            
        case DO_HOLD_STANCE:
            cmdStance = HOLD_STANCE;
            resetMovement();
            cmdHeadtype = REL_H;
            break;
            
        case NORMAL_STANCE_ABSH:
            cmdStance = NORMAL_STANCE;
            resetMovement();
            cmdHeadtype = ABS_H;
            break;
            
        case NORMAL_STANCE_RELH:
            cmdStance = NORMAL_STANCE;
            resetMovement();
            cmdHeadtype = REL_H;
            break;
            
        case SET_STANCE:
            cmdStance = intValue;
            resetMovement();
            cmdHeadtype = REL_H;
            break;
            
        case DO_TURN_KICK:
            cmdStance   = TURN_KICK;
            cmdTurnKick = intValue;
            resetMovement();
            cmdHeadtype = REL_H;
            break;
            
        case PAN_TILT:
            cout << "pan tilt" << endl;
            cmdPan      = doubleValue;
            cmdTilt     = command.value2;
            cmdHeadtype = ABS_PT; // pan first
            break;
            
        case TILT_PAN:
            cmdPan      = doubleValue;
            cmdTilt     = command.value2;
            cmdHeadtype = ABS_H; // tilt first
            break;
            
        case MOVE_COMMAND:
            cmdLeft = doubleValue;
            cmdForward = command.value2;
            cmdTurn = command.value3;
            break;
            
        case HUMAN_CONTROL:
            if (intValue==PLAYER_NUMBER) { // take control
                oldPLAYER = PLAYER;
                PLAYER    = REMOTE_ONLY;
            } else if (intValue==0) { // release control
                PLAYER = oldPLAYER;
            } else {
                break;
            }
            break;
            
        case SET_COLOUR:
            if (intValue == PLAYER_NUMBER) {
                isRedRobot = true;
            } else if (intValue==0) {
                isRedRobot = false;
            } else {
                break;
            }
            gps->SetGoals(isRedRobot);
            cout << "Robot colour changed to: ";
            if (isRedRobot) {
                cmdTailBlue = 1;
                cmdTailRed  = 2;
                cout << "Red" << endl;
            } else {
                cmdTailBlue = 2;
                cmdTailRed  = 1;
                cout << "Blue" << endl;
            }
            break;
            
        case SET_INDICATOR:
            // change the indicator specified
            intValue2 = (int) (command.value2);
            switch (intValue) {
                case 1:
                    cmdMouth = intValue2;
                    mouth    = intValue2;
                    cout << "mouth: " << intValue2 << endl;
                    break;
                case 2:
                    cmdTailH = intValue2;
                    tail_h   = intValue2;
                    cout << "tail_h: " << intValue2 << endl;
                    break;
                case 3:
                    cmdTailV = intValue2;
                    tail_v   = intValue2;
                    cout << "tail_v: " << intValue2 << endl;
                    break;
                case 4:
                    cmdEarL = intValue2;
                    ear_l   = intValue2;
                    cout << "ear_l: " << intValue2 << endl;
                    break;
                case 5:
                    cmdEarR = intValue2;
                    ear_r   = intValue2;
                    cout << "ear_r: " << intValue2 << endl;
                    break;
                case 6:
                    cmdLed1 = intValue2;
                    led1    = intValue2;
                    cout << "led1: " << intValue2 << endl;
                    break;
                case 7:
                    cmdLed2 = intValue2;
                    led2    = intValue2;
                    cout << "led2: " << intValue2 << endl;
                    break;
                case 8:
                    cmdLed3 = intValue2;
                    led3    = intValue2;
                    cout << "led3: " << intValue2 << endl;
                    break;
                case 9:
                    cmdLed4 = intValue2;
                    led4    = intValue2;
                    cout << "led4: " << intValue2 << endl;
                    break;
                case 10:
                    cmdLed5 = intValue2;
                    led5    = intValue2;
                    cout << "led5: " << intValue2 << endl;
                    break;
                case 11:
                    cmdLed6 = intValue2;
                    led6    = intValue2;
                    cout << "led6: " << intValue2 << endl;
                    break;
                case 12:
                    cmdBarLed = intValue2;
                    barLed    = intValue2;
                    cout << "barLed: " << intValue2 << endl;
                    break;
                case 13:
                    cmdTailBlue = intValue2;
                    tailBlue    = intValue2;
                    cout << "tailBlue: " << intValue2 << endl;
                    break;
                case 14:
                    cmdTailRed = intValue2;
                    tailRed    = intValue2;
                    cout << "tailRed: " << intValue2 << endl;
                    break;
            }
            break;
        case EDGE_DETECT_DEBUG:
#ifndef OFFLINE
            transmission->setSendDebug(doubleValue);
#endif // OFFLINE
            break;
            
        case BEACON_DEBUG:
            gps->sendDebug = true;
            gps->debugCounter = 0;
            gps->debugParm = intValue;
            break;
            
        case SET_BEACON_DIST_CONST:
            vision->beaconDistanceConstant = doubleValue;
            cout << "Setting Beacon Distance Constant to: " << 
            vision->beaconDistanceConstant << "\n";
            break;
			 
        case SET_CURRENTLY_CARRYING:
            if (intValue==0) {
                gps->currentlyCarrying = false;
            } else {
                gps->currentlyCarrying = true;
            }
            break;
            
        case DO_WM_LAT_TEST:
            if(intValue == PLAYER_NUMBER) {
                gps->doLatTest = true;
            }
            break;
			
		case CAL_TURN_KICK:
            intValue2 = (int) (command.value2);
            switch (intValue) {
                case 1:
					pvtkDuration = intValue2;
					break;
				case 2:
					pvtkDirection = intValue2;
					break;
				case 3:
					turnAmount = intValue2;
					break;
				case 4:
					startStep_ = intValue2;
					break;
				case 5:
					step_ = intValue2;
					break;
				case 6:
					desKickWindow = intValue2;
					break;
				default:
					break;
			}
            
        default:
            break;
    
    }
    
    command.clear();
    
    if (PLAYER==REMOTE_ONLY) {
        switch(cmdStance) {
            case NORMAL_STANCE:
                setStandParams();
                PG = 40;
                forward  = cmdForward;
                turnCCW  = cmdTurn;
                left     = cmdLeft;
                tilty    = cmdTilt;
                panx     = cmdPan;
                headtype = cmdHeadtype;
                break;
            case CHEST_PUSH:
                setChestPushParams();
                break;
            case GOALIE_KICK:
                setGoalieKickParams();
                break;
            case LIGHTNING_KICK:
                setLightningKickParams();
                break;
            case SPIN_KICK:
                setSpinKickParams();
                break;
            /*
            case TURN_KICK:
                if (*stepComplete && lockMode==ResetMode) {
                    lockMode = NoMode;
                }
                if (lockMode != NoMode && lockMode != ResetMode) {
                    if (lockMode==TurnKick) {
                        aaTurnKick();
                    }
                    if (lockMode==VariableTurnKick) {
                        aaVariableTurnKick();
                    }
                } else {
                    if (cmdTurnKick!=0) {

                        if (cmdTurnKick>0) {
                            turnDir = 1;
                            cout << "****left ";
                        } else {
                            turnDir = -1;
                            cout << "****right ";
                        }

                        cmdTurnKick = ABS(cmdTurnKick);

                        switch (cmdTurnKick) {
                        case NORMAL_TURN:
                            lockMode = TurnKick;
                            aaTurnKick();
                            cout << "normal turn kick" << endl;
                            break;
                        case VSOFT_TURN:
                            lockMode = VariableTurnKick;
                            resetVariableTurnKickCounter();
                            setVSoftVariableTurnKickType();
                            aaVariableTurnKick();
                            cout << "very soft turn kick" << endl;
                            break;
                        case SMALL_TURN:
                            lockMode = VariableTurnKick;
                            resetVariableTurnKickCounter();
                            setSmallVariableTurnKickType();
                            aaVariableTurnKick();
                            cout << "small turn kick" << endl;
                            break;
                        case MED_TURN:
                            lockMode = VariableTurnKick;
                            resetVariableTurnKickCounter();
                            setMedVariableTurnKickType();
                            aaVariableTurnKick();
                            cout << "medium turn kick" << endl;
                            break;
                        case FULL_TURN:
                            lockMode = VariableTurnKick;
                            resetVariableTurnKickCounter();
                            setFullVariableTurnKickType();
                            aaVariableTurnKick();
                            cout << "full turn kick" << endl;
                            break;
                        default:
                            break;
                        }

                        cmdTurnKick = 0;
                    }
                }
                break;
            */
            case BLOCK_STANCE:
                setBlockParams();
                forward  = cmdForward;
                turnCCW  = cmdTurn;
                left     = cmdLeft;
                tilty    = cmdTilt;
                panx     = cmdPan;
                headtype = cmdHeadtype;
                break;
            case HOLD_STANCE:
                if (cmdLeft!=0) {
                    setHoldSideWalkParams();
                    left = cmdLeft;
                } else if (cmdTurn>0) {
                    setHoldTurnParams();
                    turnCCW = 30;
                    left    = -5;
                } else if (cmdTurn<0) {
                    setHoldTurnParams();
                    turnCCW = -30;
                    left    = 5;
                } else {
                    setHoldParams();
                }
                tilty    = cmdTilt;
                panx     = cmdPan;
                headtype = cmdHeadtype;
                break;
        }

    } 

    if (cmdActiveGpsLocalise) {
        if(cmdActiveGpsLocalise == 1) {
            forward = 0;
        } else {
            turnCCW = 20;
            cmdActiveGpsLocalise--;
        } 
        
    }

}


void Behaviours::resetMovement() {
    cmdForward = 0;
    cmdLeft    = 0;
    cmdTurn    = 0;
    cmdTilt    = 0;
    cmdPan     = 0;
}

void Behaviours::saCircleAround(double step_angle, double radius) {
    // adjust for rotation centre
    radius += 7.5;

    turnCCW = step_angle;
    forward = radius - radius * cos (DEG2RAD (turnCCW));
    left = -radius * sin (DEG2RAD (turnCCW));
}

void Behaviours::setStandParams() {
    walkType = CanterWalkWT;
    forward  = 0;
    left     = 0;
    turnCCW  = 0;
    PG       = 5;
    hF       = 70;
    hB       = 110;
    hdF      = 0;
    hdB      = 0;
    ffO      = 55;
    fsO      = 15;
    bfO      = -55;
    bsO      = 15;
}

void Behaviours::setWalkParams() {
    walkType = CanterWalkWT;
    PG       = 40;
    hF       = 70;
    hB       = 110;
    hdF      = 5;
    hdB      = 25;
    ffO      = 55;
    fsO      = 15;
    bfO      = -55;
    bsO      = 15;
}

void Behaviours::setOffsetWalkParams() {
    walkType = OffsetWalkWT;
    PG       = 40;
    hF       = 70;
    hB       = 110;
    hdF      = 5;
    hdB      = 25;
    ffO      = 60;
    fsO      = 5;
    bfO      = -55;
    bsO      = 10;
}

void Behaviours::setOmniParams() {
    walkType = CanterWalkWT;
    PG       = 40;
    hF       = 70;
    hB       = 110;
    hdF      = 2;
    hdB      = 25;
    ffO      = 55;
    fsO      = 15;
    bfO      = -55;
    bsO      = 15;
}

void Behaviours::setBlockParams() {
    walkType = CanterWalkWT;
    forward  = 0;
    left     = 0;
    turnCCW  = 0;
    PG       = 5;
    hF       = 30;  // height of shoulder
    hB       = 70;  // height of hip
    hdF      = 0;   // how high front paw is lifted when walking
    hdB      = 0;   // how high back paw is lifted when walking
    ffO      = 60;  //ffO     = 100; // how far forward the front paws are
    fsO      = 110; //fsO     = 70; // how far outwards from the body the front paws are
    bfO      = -50; // how far forward the back paws are
    bsO      = 15;  // how far outwards from the body the back paws are
}

void Behaviours::setHoldParams() {
    walkType = CanterWalkWT;
    forward  = 0;
    left     = 0;
    turnCCW  = 0;
    PG       = 5;
    hF       = 70;
    hB       = 110;
    hdF      = 0;
    hdB      = 0;
    ffO      = 100;
    fsO      = -15;
    bfO      = -55;
    bsO      = 15;
}

void Behaviours::setHoldSideWalkParams() {
    walkType = CanterWalkWT;
    forward  = 0;
    left     = 0;
    turnCCW  = 0;
    PG       = 40;
    hF       = 70;
    hB       = 110;
    hdF      = 5;
    hdB      = 25;
    ffO      = 85;
    fsO      = 15;
    bfO      = -55;
    bsO      = 15;
}

void Behaviours::setHoldTurnParams() {
    setCarryParams();
}

void Behaviours::setCarryParams() {
    walkType = CanterWalkWT;
    PG       = 40;
    hF       = 70;
    hB       = 110;
    hdF      = 0;
    hdB      = 30;
    ffO      = 97.5;
    fsO      = -15;
    bfO      = -55;
    bsO      = 15;
}


void Behaviours::setRotateParams() {
    walkType = CanterWalkWT;
    forward  = 0;
    left     = 0;
    turnCCW  = -2;
    PG       = 40;
    hF       = 70;
    hB       = 110;
    hdF      = 5;
    hdB      = 30;
    ffO      = 55;
    fsO      = 15;
    bfO      = -55;
    bsO      = 15;
}

void Behaviours::printCurrentPlayer() {
    cout << "Current Player: " << PLAYER << " - ";
    switch (PLAYER) {
        case REMOTE_ONLY:
            cout << "Remote only" << endl;
            break;
        case TEAMA_GOALIE:
            cout << "Team A Goalie" << endl;
            break;
        case TEAMA_FORWARD:
            cout << "Team A Forward" << endl;
            break;
        case TEAMB_GOALIE:
            cout << "Team B Goalie" << endl;
            break;
        case TEAMB_FORWARD:
            cout << "Team B Forward" << endl;
            break;
        case BALL_COLLECTION:
            cout << "Ball Collection" << endl;
            break;
        case TEST_TURNKICK:
            cout << "Test Turnkick" << endl;
            break;
        case LOC_CHALLENGE:
            cout << "Localisation Challenge" << endl;
            break;
        case BALL_TRACK:
            cout << "Ball Tracking" << endl;
            break;
        case TEAMC_GOALIE:
            cout << "Team C Goalie" << endl;
            break;
        case MALAK_THE_AVENGER:
            cout << "Malak the Avenger" << endl;
            break;
        case TEST_ORIENT:
            cout << "Test orient" << endl;
            break;
        case TEAMD_FORWARD:
            cout << "Team D Forward" << endl;
            break;
        case OBS_CHALLENGE:
            cout << "Obstacle avoidance challenge" << endl;
            break;
        case TEAME_FORWARD:
            cout << "Team E Forward" << endl;
            break;
        case MAVERICK:
            cout << "Maverick" << endl;
            break;
        case ZANN_THE_DEFENDER:
            cout << "Zann the Defender" << endl;
            break;
        case BECKHAM_FORWARD:
            cout << "Beckham Forward" << endl;
            break;
        case PENALTY_SHOOTER:
            cout << "Penalty Shooter" << endl;
            break;
        case NEWOBS_CHALLENGE:
            cout << "New Obstacle avoidance challenge" << endl;
            break;
        default:
            cout << "Unknown Player " << PLAYER << endl;
            break;
    }
}

void Behaviours::printCurrentState(int state) {
    if (state==FindBallState) {
        cout << "FindBallState" << endl;
    } else if (state==TrackBallState) {
        cout << "TrackBallState" << endl;
    } else if (state==BlockState) {
        cout << "BlockState" << endl;
    } else if (state==ApproachState) {
        cout << "ApproachState" << endl;
    } else if (state==KickState) {
        cout << "KickState" << endl;
    } else if (state==ReturnGoalState) {
        cout << "ReturnGoalState" << endl;
    } else if (state==StandState) {
        cout << "StandState" << endl;
    } else if (state==FindBallState) {
        cout << "FindBallState" << endl;
    } else if (state==FindGoalState) {
        cout << "FindGoalState" << endl;
    } else if (state==HoldState) {
        cout << "HoldState" << endl;
    } else if (state==CarryState) {
        cout << "CarryState" << endl;
    } else if (state==BackOffState) {
        cout << "BackOffState" << endl;
    } else {
        cout << "Unknown state" << endl;
    }
}

// load the default player from a file on startup
void Behaviours::loadDefaultPlayer(const char *filename) {
    ifstream playerFile(filename);
    if (playerFile) {
        playerFile >> PLAYER_NUMBER;
        playerFile >> PLAYER;
        playerFile >> kickInix;
        playerFile >> kickIniy;
        playerFile >> kickInih;
        playerFile >> recvInix;
        playerFile >> recvIniy;
        playerFile >> recvInih;
        playerFile >> findPositionx;
        playerFile >> findPositiony;
        playerFile.close();
    }
    cout << "I am player number " << PLAYER_NUMBER << endl;
}

void Behaviours::clearEmonLog() {
#ifndef OFFLINE    
	if (emonLogDebugMsg) {
        cout << "In clearEmonLog" << endl;
    }
    ofstream outFile(EMON_LOG_FILE);
    if (emonLogDebugMsg) {
        cout << "writing to " << EMON_LOG_FILE << endl;
    }
    if (outFile.fail()) {
        cout << "Error open " << EMON_LOG_FILE << endl;
        return;
    }
    if (emonLogDebugMsg) {
        cout << "in the if statement" << endl;
    }
    outFile << EMON_LOG_CLEARED_NOTICE << endl;
    if (emonLogDebugMsg) {
        cout << "before closing the " << EMON_LOG_FILE << endl;
    }
    outFile.close();
    cout << EMON_LOG_FILE << " cleared" << endl;
    if (emonLogDebugMsg) {
        cout << "In clearEmonLog **" << endl;
    }
#endif // OFFLINE
}

// set the default indicators
void Behaviours::initIndicators() {
    cmdMouth    = 1;
    cmdTailH    = 1;
    cmdTailV    = 1;
    cmdEarL     = 1;
    cmdEarR     = 1;
    cmdLed1     = 1;
    cmdLed2     = 1;
    cmdLed3     = 1;
    cmdLed4     = 1;
    cmdLed5     = 1;
    cmdLed6     = 1;
    cmdBarLed   = inds[numINDICATORS-3];
    cmdTailBlue = inds[numINDICATORS-2];
    cmdTailRed  = inds[numINDICATORS-1];
}

double Behaviours::requiredAccuracy(double reqGoalWidth) {
    return requiredAccuracy(gps->self().pos.x, gps->self().pos.y, reqGoalWidth);
}

//returns the angle (either side) from the point specified to the goal posts
double Behaviours::requiredAccuracy(double tmx, double tmy, double reqGoalWidth) {
    double globalXoffset = FIELD_WIDTH/2 - tmx;
    double globalYoffset = FIELD_LENGTH  - tmy;
    
    double poleDistance     = sqrt(SQUARE(abs(globalXoffset)-reqGoalWidth) + SQUARE(globalYoffset));
    double globalGoalHead   = RAD2DEG( atan2(abs(globalYoffset), abs(globalXoffset)));
    double requiredAccuracy = RAD2DEG( asin(reqGoalWidth*sin(DEG2RAD(globalGoalHead))/poleDistance)  ); 
    return requiredAccuracy;
}


void Behaviours::ballSearch() {

    if(((lostBall-BALL_LOST)/INTERVAL)%2 == 0) {
        headtype = ABS_H;
        panx = QUARTER_CIRCLE - (HALF_CIRCLE/INTERVAL)*((lostBall-BALL_LOST)%INTERVAL);
        tilty = LOW_TILT;
    } else {
        headtype = ABS_H;
        panx = -QUARTER_CIRCLE + (HALF_CIRCLE/INTERVAL)*((lostBall-BALL_LOST)%INTERVAL);
        tilty = HIGH_TILT;
    }
}


// now define all the storage

namespace Behaviours {

    int PLAYER_NUMBER;
    CommonSense* sensors = NULL;
    VisualCortex* vision = NULL;
    GPS* gps = NULL;
#ifndef OFFLINE
    Vision* transmission = NULL;
#endif // OFFLINE
    
    int    *leading_leg = NULL;
    int    *currentWalkType = NULL;
    int    *shareStep_ = NULL;
    int    *currentPG = NULL;
    double *desiredPan = NULL;
    double *desiredTilt = NULL;
    
    bool* stepComplete = NULL;


    // XXX
    int mode_, modeBeforeFallen;
    RoboCupGameControlData gameState;


    // the system wide camera frame counter
    unsigned long frame;

#ifdef SENDING_DEBUG
#ifdef SENDING_WORLD_MODEL_DEBUG
    int cmdWDisplay;
#endif // SENDING_WORLD_MODEL_DEBUG
#ifdef SENDING_CPLANE
    int cmdCPlane;
    int cmdBurst;
    int burstCounter = 0;
#endif // SENDING_CPLANE
#ifdef SENDING_OPLANE
    int cmdOPlane;
#endif // SENDING_OPLANE
#endif // SENDING_DEBUG

    double cmdForward, cmdLeft, cmdTurn, cmdTilt, cmdPan;
    int cmdHeadtype;

    int cmdStationaryLocalise;
    int cmdTurnKick;
    int cmdActiveGpsLocalise;
    int cmdStance;
    int cmdMouth, cmdTailH, cmdTailV, cmdEarL, cmdEarR;
    int cmdLed1, cmdLed2, cmdLed3, cmdLed4, cmdLed5, cmdLed6;
    int cmdBarLed, cmdTailBlue, cmdTailRed;
    
    bool isRedRobot;

    int PLAYER;   // store the current behaviour
    int oldPLAYER;
    double kickInix, kickIniy, kickInih;
    double recvInix, recvIniy, recvInih;
    double findPositionx;
    double findPositiony;
    int walkType;
    double forward, left, turnCCW;
    double hF, hB, hdF, hdB, ffO, fsO, bfO, bsO;
    int PG;

    int headtype;
    double panx, tilty, nullz;
    int mouth, tail_h, tail_v, ear_l, ear_r;
    int led1, led2, led3, led4, led5, led6;
    int barLed, tailBlue, tailRed;
 
    CurObject vBall, voGoal, vtGoal;    // objects that the robot can see now
    CurObject vTeammate[NUM_TEAM_MEMBER], vOpponent[NUM_TEAM_MEMBER];
    

    CurObject wmBall, wmtGoal, wmoGoal, wmTeammate[NUM_TEAM_MEMBER], 
                wmOpponent[NUM_TEAM_MEMBER];    // objects info from the 
                                  // world model


    // wireless ball
    CurObject wlBall;
    
    // Robots with the same colour as your own team from wireless
    // This is different to your teammates telling you where they, themselves are located.
    // This is the teammate robots from other teammates sensors.
    CurObject wlTeammate[NUM_TEAM_MEMBER];
    // Robots with different colour to your team from wireless
    CurObject wlOpponent[NUM_TEAM_MEMBER];

    double hPan, hTilt, hRoll, hMouth;    // in degrees

    int inds[numINDICATORS];
    bool playerIndicatorNotSet;
    
    CommandData command;

    int actionTimer;
    
    char dataTesting [2*196*144];

    double destPan;


    double debugHf;
    double debugHdf;
    double debugWalkType;
    double debugForward;
    double debugLeft;
    double debugTurn;




    bool debugMsg;
    
    int lostBall = 0;                 // num frames ball unseen
    double lastTrackHead;         // last heading that you were tracking
                                  // (as set by saBallTracking or other source
    int lastSawBeacon;            // num frames last seen beacon
    bool isWirelessDown;          // true iff wireless is down
    double useZoidWalk;           // whether to use zoidalwalk or not.
    double dkd;                   // optimal kick direction
    double ballUnderChinTilt;     // tilt when ball under chin

    int attackMode;               // current far ball action
    int prevAttackMode;           // previous far ball action
    int attackLock;

    int grabTime;                 // num frames ball grabbed  
    int grabInterval;             // num frames disallow ball grab

    int seeBall;                  // num frames ball seen

    // find ball variables
    int lookdir;                  // direction to look first when find ball
    int turnDir;                  // direction to turn in finding ball

    int lookAroundCounter;

    double lastTilt, lastPan;
    double lastBallTilt;          // the last tilt when tracking ball
	
	bool standTall;

    // role determination variables
    int role;
    int closestRobotToBall;
    bool is3rdClosest;
    int backoffTeammateIndex; //this is not used!! was used by malak guys

    int lockMode;
    int prevLockMode;

    int findGoalTimer;
    int carryTimer;
    double goalPan;
    int holdTimer;
    int lostBallTimer;
    int backoffTimer;

    int kickTimer;
    int blockTimer;

    int state;
    double prevDistToBall;

    double ballx;
    double bally;

    int spinDirection;

// note that there is another (static) defn of teammateSeenTime in Utility.cc
    int teammateSeenTime; // number of frames teammate seen
    
    bool allowedToGoForBall;
    
    int panDirection;
    int tiltDirection;

}
