/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: UNSWSkills.cc 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 *
**/

#include "Behaviours.h"
#include "beckham.h"

#include "../vision/CommonSense.h"
#ifndef OFFLINE
#include "../vision/Vision.h"
#endif // OFFLINE
#include "../share/WirelessTypeDef.h"
#include <sstream>

static const bool delayDebugMsg = false;

static const int BURST_NUMBER            = 100;
static const int MOBILIZE_COUNT_HEAD     = 4;
static const int MOBILIZE_COUNT_BACK     = LARGE_VAL; //16; //ERIC CHANGED THIS TO DEACTIVATE BACK BUTTON
static const int DEMOBILIZE_COUNT_HEAD   = 24;
static const int DEMOBILIZE_COUNT_BACK   = LARGE_VAL; //96; //ERIC CHANGED THIS TO DEACTIVATE BACK BUTTON
static const int PRESSURE_RECOVERY_VALUE = -100;
static const int TAIL_KICKOFF            = 300000;

static const double DISTANCE_ACCURACY = 5;
static const double ANGLE_ACCURACY = 3;
static const double POSITION_VAR_ACCURACY = get95CF(20);
//static const double ANGLE_VAR_ACCURACY = get95CF(6);

static const int DISPLAY_WORLD_MODEL_ALL = 9;
static const int NEWOBS_CHALLENGE        = 23;

static const int NUM_FRAMES_TO_PAN = 139;

static int yuvSaveNumber = -1;

static AtomicAction lastAA, nextAA;
static int mouthCounter = 0;


static const int BALL_LOST = 5;

bool Behaviours::kickingOff = false;
bool Behaviours::utherKickingOff = true;

int Behaviours::malakTheAvengerKickOffTimer = 0;

bool Behaviours::beckhamKickingOff = false;
int Behaviours::beckhamKickingOffTimer = 0;

bool Behaviours::kickOffReceiverUseBallPred = false;
int Behaviours::kickOffReceiverUseBallPredTimer = 0;

#ifdef LOG_YUV
    static const int logYUVFrequency = 25;
    static int logYUVTimer = 0;
#endif

void Behaviours::UpdateMode() {

    // Detect and reset fallen down state
    if (sensors->hasFallen != NotFallenFS && mode_ != mGettingUp && *stepComplete) {
        modeBeforeFallen = mode_;
        mode_ = mGettingUp;
    } else if (mode_ == mGettingUp) {
        // Detect gotten up from fallen down
        if (sensors->hasFallen == NotFallenFS) {
            sensors->resetFallenState();
            switch (modeBeforeFallen) {
                case mDemobilized:
                case mStationaryLocalize:
                case mStartMatch:
                    mode_ = mDemobilized;
                    break;
                default:
                    mode_ = mNormal;
                    break;
            }
            //Reset();
        }
    }

    // Detect head or back press to mobilize and demobilize
    if (mode_ == mDemobilized) {
        if (sensors->switch_sensor[switch_BACK].count > MOBILIZE_COUNT_BACK ||
            sensors->press_sensor[press_HEAD_FRONT].count > MOBILIZE_COUNT_HEAD) {
            
            if (PLAYER == NEWOBS_CHALLENGE) {
                mode_ = mStartMatch;
            } else if (PLAYER == 0) {
                //save YUV plane to the stick.
                char * bflFile;
                bflFile = (char*)malloc(512);
                yuvSaveNumber ++;
                
                sprintf(bflFile, "/MS/yuv%.2d.bfl", (int)yuvSaveNumber);
                
//                sprintf(bflFile, "%s", "/MS/yuvPlaneXX.bfl");
//                const char *bflFile = "/MS/exx.bfl";
//                bflFile[13] = (char)((yuvSaveNumber%10)+48);
//                bflFile[12] = (char)((yuvSaveNumber/10)+48);

                
                
//                const char *bflFile = "/MS/blah.bfl";
                ofstream yuvFile(bflFile);
                if (yuvFile)
                {
                    /*
                    for (int i = 0; i < YUVPLANE_SIZE; i ++)
                    {
                        yuvFile << (char)(((char*)(vision->yplane))[i]);
                    }
                    */
                    for (int h = 0; h < HEIGHT; h ++)
                    {
                        //write Y
                        for (int w = 0; w < WIDTH; w ++)
                        {
                            yuvFile << (char)(((char*)(vision->yplane))[h*3*WIDTH+w]);
                        }
                        //write U
                        for (int w = 0; w < WIDTH; w ++)
                        {
                            yuvFile << (char)(((char*)(vision->yplane))[(h*3+1)*WIDTH+w]);
                        }
                        //write V
                        for (int w = 0; w < WIDTH; w ++)
                        {
                            yuvFile << (char)(((char*)(vision->yplane))[(h*3+2)*WIDTH+w]);
                        }
                        //write C
                        for (int w = 0; w < WIDTH; w ++)
                        {
                            yuvFile << (char)(((char*)(vision->cplane))[h*WIDTH+w]);
                        }
                    }
                    if (yuvSaveNumber%2)
                    {
                        leds(1,1,1,2,2,2);
                    }
                    else
                    {
                        leds(2,2,2,1,1,1);
                    }
                    yuvFile.close();
                }
                else
                {
                    leds(1,2,1,1,2,1);
                }
            } else {
                mode_ = mStationaryLocalize;
                actionTimer = NUM_FRAMES_TO_PAN;
            }
            
            sensors->switch_sensor[switch_BACK].count = PRESSURE_RECOVERY_VALUE;
            sensors->press_sensor[press_HEAD_FRONT].count = PRESSURE_RECOVERY_VALUE;
            sensors->press_sensor[press_HEAD_BACK].count = PRESSURE_RECOVERY_VALUE;
            if (sensors->sensorVal[ssTAIL_V] > TAIL_KICKOFF)
                kickingOff = true;
            else
                kickingOff = false;
        } else if (sensors->press_sensor[press_HEAD_BACK].count > MOBILIZE_COUNT_HEAD) {
            mode_ = mStartMatch;
            // Produce_Sound(0);
            sensors->switch_sensor[switch_BACK].count = PRESSURE_RECOVERY_VALUE;
            sensors->press_sensor[press_HEAD_FRONT].count = PRESSURE_RECOVERY_VALUE;
            sensors->press_sensor[press_HEAD_BACK].count = PRESSURE_RECOVERY_VALUE;
            if (sensors->sensorVal[ssTAIL_V] > TAIL_KICKOFF)
                kickingOff = true;
            else
                kickingOff = false;
        }
    } else if (mode_ == mStationaryLocalize) {
        if (actionTimer == 0) {
            mode_ = mStartMatch;
            // Produce_Sound(0);
        }
    } else if (mode_ == mStartMatch) {
        if (sensors->switch_sensor[switch_BACK].count > MOBILIZE_COUNT_BACK ||
            sensors->press_sensor[press_HEAD_FRONT].count > MOBILIZE_COUNT_HEAD) {
            cerr << "games begun" << endl;
            mode_ = mNormal;
            Reset();
            sensors->switch_sensor[switch_BACK].count = PRESSURE_RECOVERY_VALUE;
            sensors->press_sensor[press_HEAD_FRONT].count = PRESSURE_RECOVERY_VALUE;
        }
    } 
	
	else if (sensors->switch_sensor[switch_BACK].count > DEMOBILIZE_COUNT_BACK ||
        sensors->press_sensor[press_HEAD_FRONT].count > DEMOBILIZE_COUNT_HEAD) {
        // Produce_Sound(14);
        mode_ = mDemobilized;
        sensors->switch_sensor[switch_BACK].count = PRESSURE_RECOVERY_VALUE;
        sensors->press_sensor[press_HEAD_FRONT].count = PRESSURE_RECOVERY_VALUE;
        Reset();
    } 
	
	else if(mode_ == mKickoffPos) {
	
		double distDiff = sqrt(SQUARE(gps->self().pos.x-kickInix)
			+ SQUARE(gps->self().pos.y-kickIniy));
		double angDiff = ABS(gps->self().h - kickInih);
		
		if((distDiff < DISTANCE_ACCURACY && angDiff < ANGLE_ACCURACY &&
		   gps->self().posVar < POSITION_VAR_ACCURACY) ||
		   (sensors->press_sensor[press_HEAD_BACK].count > MOBILIZE_COUNT_HEAD)) {
			setStandParams();
			makeParaWalk(walkType, forward, left, turnCCW, PG, 
                         hF, hB, hdF, hdB, ffO, fsO, bfO, bsO);
			//actionTimer = NUM_FRAMES_TO_PAN;
            //mode_ = mStationaryLocalize;
			mode_ = mStartMatch;
		}
		
	} else if(mode_ == mReceivePos) {
	
		double distDiff = sqrt(SQUARE(gps->self().pos.x-recvInix)
			+ SQUARE(gps->self().pos.y-recvIniy));
		double angDiff = ABS(gps->self().h - recvInih);
		
		if((distDiff < DISTANCE_ACCURACY && angDiff < ANGLE_ACCURACY &&
		   gps->self().posVar < POSITION_VAR_ACCURACY) ||
		   (sensors->press_sensor[press_HEAD_BACK].count > MOBILIZE_COUNT_HEAD)) {
			setStandParams();
			makeParaWalk(walkType, forward, left, turnCCW, PG, 
                         hF, hB, hdF, hdB, ffO, fsO, bfO, bsO);
			//actionTimer = NUM_FRAMES_TO_PAN;
            //mode_ = mStationaryLocalize;
			mode_ = mStartMatch;
		}
		
	}
}


void Behaviours::NextAction() {
    frame++;
    UpdateMode();
    initialize();
	
	// keep track of how long we have lost the ball for
	if (gps->canSee(vobBall)) {        
        lostBall = 0;        
    } else {
        lostBall++;            
    }

    switch (mode_) {
        case mDemobilized:
            // show battery status
            BatteryLEDS();
            
            // sit up
            makeParaWalk(CanterWalkWT, 0, 0, 0, 60, 73, 107, 0, 0, 50, 15, -50, 15, ABS_H, 0, 0, 0);
            // makeParaWalk(0,2,0,0,60,73,107,0,0,50,15,-50,15,ABS_H,0,0);
            break;

        case mGettingUp:
            if (sensors->hasFallen == FallenFrontFS)
                makeParaWalk(GetupFrontWT,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0);
            else if (sensors->hasFallen == FallenRearFS)
                makeParaWalk(GetupRearWT,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0);
            else if (sensors->hasFallen == FallenSideFS)
                makeParaWalk(GetupSideWT,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0);
            break;

        case mStationaryLocalize:
            // show battery status
            BatteryLEDS();
			
            StationaryLocalize();
            break;
			
		case mKickoffPos:
            // show battery status
            BatteryLEDS();
			
			kickingOff = (sqrt(SQUARE(kickInix-FIELD_WIDTH/2) + SQUARE(kickIniy-FIELD_LENGTH/2)) < 20);
			smartSetBeacon();
        	smartActiveGpsLocalise();
			//saGoToTargetFacingHeading(kickInix, kickIniy, kickInih);
			Beckham::doFastGoToPointHeading(kickInix, kickIniy, kickIniy);
			setWalkParams();
			makeParaWalk (walkType, forward, left, turnCCW, PG, 
                          hF, hB, hdF, hdB, ffO, fsO, bfO, bsO);
			makeHead(headtype, panx, tilty, nullz);
			break;
			
		case mReceivePos:
            // show battery status
            BatteryLEDS();
			
			kickingOff = false;
			smartSetBeacon();
        	smartActiveGpsLocalise();
			//saGoToTargetFacingHeading(recvInix, recvIniy, recvInih);
			Beckham::doFastGoToPointHeading(recvInix, recvIniy, recvInih);
			setWalkParams();
			makeParaWalk (walkType, forward, left, turnCCW, PG, 
                          hF, hB, hdF, hdB, ffO, fsO, bfO, bsO);
			makeHead(headtype, panx, tilty, nullz);
			break;

        case mStartMatch:
			
			// do decideNextAction so the behaviours is being run, later setStandParams
			// will be called so robot doesn't move, behaviour still 
			DecideNextAction();
		
            // show battery status
            BatteryLEDS();
			            
            // set walk params
            setStandParams();
			if(standTall) hF = 85;
            makeParaWalk(walkType, forward, left, turnCCW, PG,
                hF, hB, hdF, hdB, ffO, fsO, bfO, bsO, headtype, panx, tilty, nullz);

            // reset kickoff timer
            utherKickingOff = true;
            malakTheAvengerKickOffTimer = 0;
            break;
            
        default:
            DecideNextAction();
    }
	
	// send information for other dogs
#ifndef OFFLINE
	if(mode_ != mDemobilized)
    	transmission->shareEnvironment();
#endif // OFFLINE
	
	// send any debug information
#ifdef SENDING_DEBUG
#ifdef SENDING_CPLANE
    // send cplane data
    if (cmdCPlane == PLAYER_NUMBER) {
#ifndef OFFLINE
        transmission->sendCompressedCPlane();
#endif // OFFLINE
    }
    // send burst data
    if (cmdBurst == PLAYER_NUMBER && Behaviours::burstCounter > 0 ) {
#ifndef OFFLINE
        transmission->sendCompressedCPlane();
        burstCounter--;
#endif // OFFLINE
    }
#endif // SENDING_BURST, SENDING_CPLANE

#ifdef SENDING_OPLANE
    // send oplane data
    if (cmdOPlane == PLAYER_NUMBER) {
#ifndef OFFLINE
        transmission->sendOPlane();
#endif // OFFLINE
    }
#endif // SENDING_OPLANE

#ifdef SENDING_WORLD_MODEL_DEBUG
    // send world model stuff
    if (cmdWDisplay == PLAYER_NUMBER || cmdWDisplay == DISPLAY_WORLD_MODEL_ALL) {
#ifndef OFFLINE
	    transmission->displayEnvironment();
#endif // OFFLINE
    }
#endif // SENDING_WORLD_MODEL_DEBUG

#ifdef LOG_YUV
    if (logYUVTimer == logYUVFrequency) {
        transmission->sendYUVPlane();
        logYUVTimer = 0;            
    } else {
        logYUVTimer++;            
    }            
#endif
    
#endif // SENDING_DEBUG

#ifndef OFFLINE
	if(transmission->gotTMWM) {
		// turn bar on
		inds[numINDICATORS - 3] = 2;
		transmission->gotTMWM = 0;
	} else {
		// turn bar off
		inds[numINDICATORS - 3] = 1;
	}
	
	// if there are three or more teammates with valid world models turn bar on
	int tm = 0;
	//cout << "friends = {";
	for(int i=0 ; i<NUM_TEAM_MEMBER ; i++) {
		if(gps->tmObj(i).counter > 0) {
			//cout << i+1 << " ";
			tm++;
		}
	}
	//cout << "}" << endl;
	if(tm >=3) {
		inds[numINDICATORS - 3] = 2;
		inds[1] = TAIL_CENTRED;
	} else if(tm > 0) {
		inds[1] = TAIL_LEFT;
	} else
		inds[1] = TAIL_RIGHT;

    transmission->sendAA(nextAA);
    transmission->sendIndicators(inds, sizeof(inds));
#endif // OFFLINE
}

void Behaviours::BatteryLEDS() {

#ifndef OFFLINE
    OStatus result;
    OPowerStatus currentStatus;

    result = OPENR::GetPowerStatus(&currentStatus);
    if(result == oSUCCESS) {
        word charge = currentStatus.remainingCapacity;
        inds[5]  = 1;
        inds[6]  = 1;
        inds[7]  = 1;
        inds[8]  = 1;
        inds[9]  = 1;
        inds[10] = 1;
        if(charge > 100/7)
            inds[5] = 2;
        if(charge > 200/7)
            inds[8] = 2;
        if(charge > 300/7)
            inds[6] = 2;
        if(charge > 400/7)
            inds[9] = 2;
        if(charge > 500/7)
            inds[7] = 2;
        if(charge > 600/7)
            inds[10] = 2;
		// took this out as bar led now used to indicate if wireless is up
        /*if(currentStatus.robotStatus & orsbBATTERY_CAPACITY_FULL)
            inds[11] = 2;*/
        if(currentStatus.robotStatus & orsbBATTERY_CAPACITY_LOW)
            inds[0] = MOUTH_OPEN;
        else
            inds[0] = MOUTH_CLOSED;
    } else {
        cout << "Error getting power status!" << endl;
    }
#endif // OFFLINE

}

void Behaviours::StationaryLocalize() {
    if (actionTimer > 0) {
        if (ABS (hPan - destPan) <= 10) {
            destPan = (destPan < 0) ? 85 : -85;
        }
        headtype = REL_H;
        panx = (destPan > 0) ? 5 : -5;
        tilty = (10 - hTilt) / 5.0;
        makeHead (headtype, panx, tilty, 0);
        actionTimer--;
    }
}

void Behaviours::makeStop() {
    // stop all actions in all networks
    nextAA.cmd = aaStop;
    defaultInit();
}

void Behaviours::makeStill(void) {
    nextAA.cmd = aaStill;
    defaultInit();
}

void Behaviours::makeParaWalk(double forward, double left, double tccw) {
    nextAA.cmd = aaParaWalk;
    nextAA.walkType = NormalWalkWT;
    nextAA.Forward = forward;    //save distance 6cm
    nextAA.Left = left;        // 4cm
    nextAA.turnCCW = tccw;    //36deg
    nextAA.speed = 80;
    nextAA.hF = 100;
    nextAA.hB = 120;
    nextAA.hdF = 15;
    nextAA.hdB = 18;
    nextAA.ffO = 25;
    nextAA.fsO = 20;
    nextAA.bfO = -25;
    nextAA.bsO = 20;
    nextAA.headType = NULL_H;
    defaultInit();
}

void Behaviours::makeParaWalk(int walkType, double Forward, double Left,
                              double turnCCW, int speed, double hF,
                              double hB, double hdF, double hdB,
                              double ffO, double fsO, double bfO, double bsO) {
    nextAA.cmd = aaParaWalk;
    nextAA.walkType = walkType;
    nextAA.Forward = Forward;
    nextAA.Left = Left;
    nextAA.turnCCW = turnCCW;
    nextAA.speed = speed;
    nextAA.hF = hF;
    nextAA.hB = hB;
    nextAA.hdF = hdF;
    nextAA.hdB = hdB;
    nextAA.ffO = ffO;
    nextAA.fsO = fsO;
    nextAA.bfO = bfO;
    nextAA.bsO = bsO;
    nextAA.headType = NULL_H;
    defaultInit();
}

void Behaviours::makeParaWalk(int walkType, double Forward, double Left,
                              double turnCCW, int speed, double hF,
                              double hB, double hdF, double hdB,
                              double ffO, double fsO, double bfO,
                              double bsO, int headType, double panx, double tilty, double nullz) {
    nextAA.cmd = aaParaWalk;
    nextAA.walkType = walkType;
    nextAA.Forward = Forward;
    nextAA.Left = Left;
    nextAA.turnCCW = turnCCW;
    nextAA.speed = speed;
    nextAA.hF = hF;
    nextAA.hB = hB;
    nextAA.hdF = hdF;
    nextAA.hdB = hdB;
    nextAA.ffO = ffO;
    nextAA.fsO = fsO;
    nextAA.bfO = bfO;
    nextAA.bsO = bsO;
    nextAA.headType = headType;
    nextAA.tilty = tilty;
    nextAA.panx = panx;
    nextAA.nullz = nullz;
    //  nextAA.mouth = mouth;
    defaultInit();
}

void Behaviours::makeHead(int headType, double panx, double tilty, double nullz) {
    nextAA.cmd = aaParaWalk;
    //  nextAA.walkType =1000;
    nextAA.headType = headType;
    nextAA.panx = panx;
    nextAA.tilty = tilty;
    nextAA.nullz = nullz;
    //  nextAA.mouth = 0;
    defaultInit();
}

void Behaviours::defaultInit(void) {
    // common initialization for all commands
    nextAA.step = 0;
    nextAA.finished = false;
}

double Behaviours::microRad2Deg(double rad) {
    return rad * 180.0 / (1000000.0 * pi);
}

void Behaviours::ExtractMotionFactors(AtomicAction * currentAA,
                                      AtomicAction * lastAA, long *dx,
                                      long *dy, long *dh, long *cf) {
    long numMoves = currentAA->param2;

    if (currentAA->cmd != lastAA->cmd) {
        numMoves -= 2;
        if (numMoves < 0) {
            numMoves = 0;
        }
    }

    *dx = 0;
    *dy = 0;
    *dh = 0;
    *cf = 990;
}

// given an object at (x,y), find its heading,distance
// relative to the robot(rx,ry,rh)
void Behaviours::findAngleDist(double x, double y, double *head, double *dist) {

	const WMObj& r = gps->self();
    *dist = sqrt ((r.pos.x - x) * (r.pos.x - x) + (r.pos.y - y) * (r.pos.y - y));
    *head = asin (abs (r.pos.y - y) / *dist) * 180 / pi;
    if (y >= r.pos.y) {
        if (x <= r.pos.x)
            *head = 180 - *head;
    } else {
        if (x >= r.pos.x)
            *head *= -1;
        else
            *head -= 180;
    }
    *head -= r.h;
    while (*head > 180) {
        *head -= 360;
    }
    while (*head <= -180) {
        *head += 360;
    }
}

// cleaned up code by using operator= and several
// object operations to make it simple
void Behaviours::initialize(void) {

    // Setting player (team and goalie) indicators
    if (playerIndicatorNotSet) {
        // Orange indicator for Red robots
        inds[numINDICATORS - 1] = gps->targetGoal == vobYellowGoal ? 1 : 2;
        // Blue indidcator for Blue robots
        inds[numINDICATORS - 2] = gps->targetGoal == vobYellowGoal ? 2 : 1;
        // Indicator for Goalie
        inds[numINDICATORS - 3] = 1;
        playerIndicatorNotSet = false;
    }
                                        // ball from current image
    if (vision->vob[vobBall].cf == 0) {
		vBall.vob = &(vision->vob[vobBall]);
        vBall.reset();
    } else {
        vBall = vision->vob[vobBall];
        
		const WMObj& r = gps->self();
        vBall.calcXY (r.pos.x, r.pos.y, r.h);
    }

//!!ERIC MAKE THIS COME BACK LATER
//!!ERIC MAKE THIS COME BACK LATER
//!!ERIC MAKE THIS COME BACK LATER
/*
    // ball from world model
    // can't use x and y in wmobj, because they are not reversed for the blue team
    wmBall = wmobj[vobBall];
    wmBall.calcXY (rx, ry, rh);

    //wout << "gps best ball x, y, h, d " << wlBestBall.x << "," << wlBestBall.y << "," << wlBestBall.h << "," << wlBestBall.d << "\n";
    wlBall = wlBestBall;
    wlBall.calcXY (rx, ry, rh);

    // tGoal -targetGoal- from the world model
    wmtGoal.x = FIELD_WIDTH / 2;
    wmtGoal.y = FIELD_LENGTH + GOAL_LENGTH / 2;    // these are fixed
    //wmtGoal.calcHD(rx,ry);
    findAngleDist (wmtGoal.x, wmtGoal.y, &wmtGoal.h, &wmtGoal.d);
    wmtGoal.cf = rcf;
    wmtGoal.var = rvar;

    // oGoal -ownGoal- from the world model
    wmoGoal.x = FIELD_WIDTH / 2;
    wmoGoal.y = -GOAL_LENGTH / 2;
    //wmoGoal.calcHD(rx,ry);
    findAngleDist (wmoGoal.x, wmoGoal.y, &wmoGoal.h, &wmoGoal.d);
    wmoGoal.cf = rcf;
    wmoGoal.var = rvar;

*/

    // tGoal from current image
    if (vision->vob[gps->targetGoal].cf == 0) {
        vtGoal.reset();
    } else {
        // tGoal from current image
        vtGoal = vision->vob[gps->targetGoal];
        vtGoal.x = FIELD_WIDTH / 2;
        vtGoal.y = FIELD_LENGTH + GOAL_LENGTH / 2;
    }

    // oGoal from current image
    if (vision->vob[gps->ownGoal].cf == 0) {
        voGoal.reset();
    } else {
        voGoal = vision->vob[gps->ownGoal];
        voGoal.x = FIELD_WIDTH / 2;
        voGoal.y = -GOAL_LENGTH / 2;
    }

    // convert head sensors to degrees
    hTilt  = microRad2Deg (sensors->sensorVal[ssHEAD_PITCH]);
    hPan   = microRad2Deg (sensors->sensorVal[ssHEAD_YAW]);
    hRoll  = microRad2Deg (sensors->sensorVal[ssHEAD_ROLL]);
    hMouth = microRad2Deg (sensors->sensorVal[ssMOUTH]);
	
    // other robots
    int wmOpponentBase = vobBlueDog, wmTeamBase = vobRedDog;
    int wlOpponentBase = NUM_TEAM_MEMBER, wlTeamBase = 0;
    if (gps->targetGoal == vobYellowGoal) {
        wmOpponentBase = vobRedDog, wmTeamBase = vobBlueDog;
        wlOpponentBase = 0, wlTeamBase = NUM_TEAM_MEMBER;
    }
    int vOpponentBase = wmOpponentBase, vTeamBase = wmTeamBase;

    // this is needed for visual backoff!
    for (int i = 0; i < NUM_TEAM_MEMBER; i++) {
        // team members from vision
        vTeammate[i] = vision->vob[vTeamBase++];
        vTeammate[i].calcXY (gps->self().pos.x, gps->self().pos.y, gps->self().getHackedH());
        // opponent members from vision
        vOpponent[i] = vision->vob[vOpponentBase++];
        vOpponent[i].calcXY (gps->self().pos.x, gps->self().pos.y, gps->self().getHackedH());
    }            

    //!!ERIC MAKE THIS COME BACK LATER
    //!!ERIC MAKE THIS COME BACK LATER
    //!!ERIC MAKE THIS COME BACK LATER
        /*
        for (int i = 0; i < NUM_TEAM_MEMBER; i++) {
        // team members as told by them
        teammate[i] = wlTeam[i];
        teammate[i].calcXY (rx, ry, rh);

        // team members from vision
        vTeammate[i] = vision->vob[vTeamBase++];
        vTeammate[i].calcXY (rx, ry, rh);

        // teammates from world model
        wmTeammate[i] = wmobj[wmTeamBase++];
        wmTeammate[i].calcXY (rx, ry, rh);

        // teammates from other ppl and your own world model
        wlTeammate[i] = wlRobots[wlTeamBase++];
        wlTeammate[i].calcXY (rx, ry, rh);


        // opponent from world model
        wmOpponent[i] = wmobj[wmOpponentBase++];
        wmOpponent[i].calcXY (rx, ry, rh);

        // opponent from other ppl and your own world model
        wlOpponent[i] = wlRobots[wlOpponentBase++];
        wlOpponent[i].calcXY (rx, ry, rh);
    }
        */
		
	standTall = false;
}

// Musician

  /*
     void UNSWSkills::SetUpSounds(void) {
     sounds[0].Init(220.00, 2048);   //A3
     sounds[1].Init(246.94, 2048);   //B3
     sounds[2].Init(261.63, 2048);   //C4
     sounds[3].Init(293.66, 2048);   //D4
     sounds[4].Init(329.63, 2048);   //E4
     sounds[5].Init(349.23, 2048);   //F4
     sounds[6].Init(392.00, 2048);   //G4
     sounds[7].Init(440.00, 2048);   //A4
     sounds[8].Init(493.88, 2048);   //B4
     sounds[9].Init(523.25, 2048);   //C5
     sounds[10].Init(587.33, 2048);  //D5
     sounds[11].Init(659.26, 2048);  //E5
     sounds[12].Init(698.46, 2048);  //F5
     sounds[13].Init(783.99, 2048);  //G5
     sounds[14].Init(880.00, 2048);  //A5
     }
   */

bool Behaviours::Produce_Sound(unsigned int index) {
    //  cout << "Sound request: " << index << endl;
    /*
     if (index >= NUM_SOUNDS)
     return false;

     bool rtn = Play(&sounds[index], 1);
     if (rtn)
     Play();
     return rtn;
    */
    return true;
}

bool Behaviours::Produce_Signal(unsigned int tone) {
    //cout << "Executing sound pattern" << endl;
    /*
     bool rtn = Play(tone);
     if (rtn)
     Play();
    */
    return true;
}

void Behaviours::makeNormalWalk(double Forward, double Left,
                                double turnCCW, int headType, double panx,
                                double tilty, double nullz) {
    makeParaWalk(CanterWalkWT, Forward, Left, turnCCW, 60, 73, 107, 5, 25, 50, 15, -50, 15, headType, panx, tilty, nullz);
}

void Behaviours::makeEdgeWalk(double Forward, double Left, double turnCCW,
                              int headType, double panx, double tilty, double nullz) {
    makeParaWalk(NormalWalkWT, Forward, Left, turnCCW, 60, 75, 110, 5, 25, 60, 5, -50, 5, headType, panx, tilty, nullz);
}

void Behaviours::makeKickWalk() {
    makeParaWalk(FwdKickWT, 0, 0, 0, 60, 73, 107, 5, 25, 50, 15, -50, 15, ABS_H, 0, 0, 0);
}

void Behaviours::makeIndicators(int mouth, int tail_h, int tail_v,
                                int ear_l, int ear_r, int led1, int led2,
                                int led3, int led4, int led5, int led6) {
    makeIndicators(mouth, tail_h, tail_v, ear_l, ear_r, 
                   led1, led2, led3, led4, led5, led6,
                   inds[11], inds[12], inds[13]);
}

void Behaviours::makeIndicators(int mouth, int tail_h, int tail_v,
                                int ear_l, int ear_r, int led1, int led2,
                                int led3, int led4, int led5, int led6,
                                int barLed, int tailBlue, int tailRed) {
    inds[0]  = mouth;
    inds[1]  = tail_h;
    inds[2]  = tail_v;
    inds[3]  = ear_l;
    inds[4]  = ear_r;
    inds[5]  = led1;
    inds[6]  = led2;
    inds[7]  = led3;
    inds[8]  = led4;
    inds[9]  = led5;
    inds[10] = led6;
    inds[11] = barLed;
    inds[12] = tailBlue;
    inds[13] = tailRed;
}

void Behaviours::wreceive(int from, int type, int size, const void *data) {
    switch (type) {
    case WL_SHARE_OBJ:
        //received a worldmodel packet from teammate
        //gps->processShareInfo (from, (const WMShareObj *) data);
        break;

    case WL_NNMC:
        vision->wlLoadNNMC ((const unsigned char *) data, size);
        break;

    case WL_COMMAND:
        /*strcpy (cmdName, (const char *) data);

        if (delayDebugMsg) {
            cout << "UNSWSkills.cc name = " << cmdName << endl;
        }*/

        break;

    case WL_VALUE:
        /*strcpy (cmdValue, (const char *) data);
        // check for mode modification
        if (strcmp (cmdName, "ip") == 0) {
            PLAYER_NUMBER = atoi(cmdValue);
            cmdName[0] = '\0';
            cmdValue[0] = '\0';
            cout << "UNSWSkills.cc My Player Number = " << PLAYER_NUMBER << endl;
            transmission->sendPlayerNumber(PLAYER_NUMBER);

        } else if (strcmp(cmdName, "ko") == 0) {
            int value = atoi(cmdValue);                            
            if (value == 1) {
                utherKickingOff = true;
            } else {
                utherKickingOff = false;
            } 
        } else if (strcmp(cmdName, "mo")==0 || strcmp(cmdName, "mode")==0) {
            mode_       = atoi (cmdValue);
            cmdName[0]  = '\0';
            cmdValue[0] = '\0';

            if (mode_==mStationaryLocalize) {
                actionTimer = NUM_FRAMES_TO_PAN;
            }
        }

        if (delayDebugMsg) {
            cout << "UNSWSkills.cc value = " << cmdValue << endl;
        }*/

        break;
    }
}

void Behaviours::processCommand(const CommandData& cmd) {

	if(cmd.type == ROBOCOMMANDER_DATA) {
    	if (cmd.RC_COMMAND == UTHER_KICKING_OFF) {
/*            
        	int kickOff = (int)(cmd.value);
        	if (kickOff) {
        	    utherKickingOff = true;
        	} else {
        	    utherKickingOff = false;
        	}
*/        
    	} else if (cmd.RC_COMMAND == MODE_CHANGE) {
        	mode_       = (int)(cmd.value);

        	if (mode_==mStationaryLocalize) {
        	    actionTimer = NUM_FRAMES_TO_PAN;
        	}
        	command.clear();
        
    	} 
#ifdef SENDING_DEBUG
#ifdef SENDING_CPLANE
		// setting the sending of debug, moved from Behaviours.cc
        else if(cmd.RC_COMMAND == CPLANE_SEND) {
            cmdCPlane = (int)(cmd.value);
        }
		// setting the sending of debug
        else if(cmd.RC_COMMAND == BURST_MODE) {
            cmdBurst = (int)(cmd.value);
            burstCounter = BURST_NUMBER;
        }
#endif // SENDING_CPLANE
            
#ifdef SENDING_WORLD_MODEL_DEBUG
        else if(cmd.RC_COMMAND == WM_SEND) {
            cmdWDisplay = (int)(cmd.value);
        }
#endif // SENDING_WORLD_MODEL_DEBUG
            
#ifdef SENDING_YUVPLANE
        else if(cmd.RC_COMMAND == YUVPLANE_SEND) {
#ifndef OFFLINE
            transmission->sendYUVPlane();
#endif // OFFLINE
        }
#endif // SENDING_YUVPLANE
            
#ifdef SENDING_OPLANE
        else if(cmd.RC_COMMAND == OPLANE_SEND) {
            cmdOPlane = (int)(cmd.value);
        }
#endif // SENDING_OPLANE
#endif // SENDING_DEBUG
		else
        	command = cmd;
	} else { // cmd.type == GAMECONTROLLER_DATA
	
		if((cmd.GC_DATA.state == ROBOCUP_STATE_FINISH) &&
		   (cmd.GC_DATA.ownScore > gameState.ownScore)) {
		   
		   lockMode = CelebrationLock;
		   
		}

		gameState = cmd.GC_DATA;
		switch(gameState.state) {
			
			case ROBOCUP_STATE_INITIAL:
				gps->setPenalized(false);
				changeMode(mStationaryLocalize);
                actionTimer = NUM_FRAMES_TO_PAN;
				break;
			
			case ROBOCUP_STATE_READY:
				gps->setPenalized(false);
				//changeMode(mStationaryLocalize);
                //actionTimer = NUM_FRAMES_TO_PAN; 
				changeMode(mStartMatch);
				

                // need toreset beckhamKickingOff for the third forward to pan
                if (gameState.kickoff == ROBOCUP_KICKOFF_OWN) {

                    beckhamKickingOff = true;       
                    beckhamKickingOffTimer = 0;

                    kickOffReceiverUseBallPred = false;
                    kickOffReceiverUseBallPredTimer = 0;
                } else {
                    beckhamKickingOff = false;
                }
				break;
				
			case ROBOCUP_STATE_PLAYING:
				gps->setPenalized(false);
				changeMode(mNormal);
                
                // needed for kick-off timer
                if (gameState.kickoff == ROBOCUP_KICKOFF_OWN) {

                    beckhamKickingOff = true;       
                    beckhamKickingOffTimer = 0;

                    kickOffReceiverUseBallPred = false;
                    kickOffReceiverUseBallPredTimer = 0;
                } else {
                    beckhamKickingOff = false;
                }    
				break;
			
			case ROBOCUP_STATE_PENALIZED:
				setPenaltyPos();
				gps->setPenalized(true);
				changeMode(mDemobilized);
				break;
			
			case ROBOCUP_STATE_FINAL:
				changeMode(mDemobilized);
				break;
			
			
		}
	
	}

}

void Behaviours::setPenaltyPos() {
	
	double angle = atan(180.0/(double)FIELD_WIDTH);
	double sinA = sin(DEG2RAD(angle));
	double cosA = cos(DEG2RAD(angle));
	double ev1 = get95CF(FIELD_WIDTH/2);
	double ev2 = get95CF(10);
	double ev3 = get95CF(10);

	// the eigenvectors for the first and third eigenvalues, in terms
	// of x and h
	double x1 = cosA;
	double h1 = sinA;
	double x2 = -sinA;
	double h2 = cosA;

	MMatrix3 newC(0);
	newC(0,0) = ev1*x1*x1 + ev3*x2*x2;
	newC(2,0) = ev1*x1*h1 + ev3*x2*h2;
	newC(0,2) = newC(1,0);
	newC(2,2) = ev1*h1*h1 + ev3*h2*h2;
	newC(1,1) = ev2;

	gps->GPSSetVals(FIELD_WIDTH/2,FIELD_LENGTH/2,QUARTER_CIRCLE,newC);
}

