/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: Common.cc 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
**/

#include <iostream>
#include "Common.h"

using namespace std;

int byte2UInt (byte b) {
    int i, s;

    s = (b >> 7) & 0x01;         // save the sign bit
    b &= 0x7f;                   // strip the sign bit
    i = (((int) b) | (s << 7));  // reassemble number

    return i;
}

double NormalizeAngle_0_360 (double angle) {
    while (angle < 0)
        angle += 360;
    while (angle >= 360)
        angle -= 360;
    return angle;
}

double NormalizeAngle_180 (double angle) {
    while (angle < -180)
        angle += 360;
    while (angle >= 180)
        angle -= 360;
    return angle;
}

double NormalizeAngle_PI (double angle) {
    while (angle < -M_PI)
        angle += 2 * M_PI;
    while (angle >= M_PI)
        angle -= 2 * M_PI;
    return angle;
}

double NormalizeAngle_m90_270 (double angle) {
    while (angle < -QUARTER_CIRCLE)
        angle += FULL_CIRCLE;
    while (angle >= (3*QUARTER_CIRCLE))
        angle -= FULL_CIRCLE;
    return angle;
}

double NormalizeAngle_90_450 (double angle) {
    while (angle < QUARTER_CIRCLE)
        angle += FULL_CIRCLE;
    while (angle >= (5*QUARTER_CIRCLE))
        angle -= FULL_CIRCLE;
    return angle;
}

void FieldClipping (double &x, double &y) {
    // left boundary
    if (x < WALL_THICKNESS)
        x = WALL_THICKNESS;

    // right boundary
    else if (x > FIELD_WIDTH - WALL_THICKNESS)
        x = FIELD_WIDTH - WALL_THICKNESS;

    // inside own goal
    if (y < -(GOAL_LENGTH / 2.0) && ABS (x - FIELD_WIDTH / 2.0) < GOAL_WIDTH / 2.0)
        y = -(GOAL_LENGTH / 2.0);

    // inside target goal - clip to field to make dogs shoot at goal rather than backwards
    else if (y > FIELD_LENGTH
             && ABS (x - FIELD_WIDTH / 2.0) < GOAL_WIDTH / 2.0)
        y = FIELD_LENGTH;

    // own goal boundary
    if (y < WALL_THICKNESS && ABS (x - FIELD_WIDTH / 2.0) > GOAL_WIDTH / 2.0)
        y = WALL_THICKNESS;

    // taget goal boundary
    else if (y > FIELD_LENGTH - WALL_THICKNESS && ABS (x - FIELD_WIDTH / 2.0) > GOAL_WIDTH / 2.0)
        y = FIELD_LENGTH - WALL_THICKNESS;

    // Now adjust for the corners

    // own goal left
    double d = 0;
    if (y < CORNER_WIDTH - WALL_THICKNESS
        && x < CORNER_WIDTH - WALL_THICKNESS && x + y - CORNER_WIDTH < 0) {
        d = ABS (x + y - CORNER_WIDTH) / 2.0;
        y += d;
        x += d;
    }
    // own goal right
    else if (y < CORNER_WIDTH - WALL_THICKNESS
             && x > FIELD_WIDTH - CORNER_WIDTH + WALL_THICKNESS
             && x - y - FIELD_WIDTH + CORNER_WIDTH < 0) {
        d = ABS (x - y - FIELD_WIDTH + CORNER_WIDTH) / 2.0;
        y += d;
        x -= d;
    }
    // target goal left
    else if (y > FIELD_LENGTH - CORNER_WIDTH + WALL_THICKNESS
             && x < CORNER_WIDTH - WALL_THICKNESS
             && x - y + FIELD_LENGTH - CORNER_WIDTH > 0) {
        d = ABS (x - y + FIELD_LENGTH - CORNER_WIDTH) / 2.0;
        y -= d;
        x += d;
    }
    // target goal right
    else if (y > FIELD_LENGTH - CORNER_WIDTH + WALL_THICKNESS
             && x > FIELD_WIDTH - CORNER_WIDTH + WALL_THICKNESS
             && x + y - FIELD_WIDTH - FIELD_LENGTH + WALL_THICKNESS > 0) {
        d = ABS (x + y - FIELD_WIDTH - FIELD_LENGTH + WALL_THICKNESS) / 2.0;
        y -= d;
        x -= d;
    }
}

bool IsInField (double x, double y) {
    double x_ = x, y_ = y;
    FieldClipping (x_, y_);
    return (x == x_) && (y == y_);
}

bool cutVertical (double cX, double cY, double pX, double pY, double Xval, double *resY) {
    if (cX == pX)
        return false;
    if ((Xval - cX) * (Xval - pX) > 0)
        return false;

    *resY = (Xval - cX) * (cY - pY) / (cX - pX) + cY;
    //  cout << " Vertical " << Xval << " " <<pX<<" "<<pY<< " " << *resY << endl;
    if ((*resY - WALL_THICKNESS) * (*resY - FIELD_LENGTH + WALL_THICKNESS) > 0)
        return false;
    else
        return true;
}

bool cutHorizontal (double cX, double cY, double pX, double pY, double Yval, double *resX) {
    if (cY == pY)
        return false;
    if ((Yval - cY) * (Yval - pY) > 0)
        return false;

    *resX = (Yval - cY) * (cX - pX) / (cY - pY) + cX;
    //  cout << " Horizontal " << Yval << " " <<pX<<" "<<pY<< " " << *resX 
    //       << endl;
    if ((*resX - WALL_THICKNESS) * (*resX - FIELD_WIDTH + WALL_THICKNESS) > 0)
        return false;
    else
        return true;
}

void angleClipping (double cX, double cY, double *pX, double *pY) {
    double x, y;
    //  cout << "angleclipping " << cX<<" "<<cY<<" "<<*pX<<" "<<*pY<<endl;
    if (cX > WALL_THICKNESS
        && cutVertical (cX, cY, *pX, *pY, WALL_THICKNESS, &y) == true) {

        // cout << "Angle Clipping " << (long) *pX << " "<<(long) *pY;
        *pX = WALL_THICKNESS;
        *pY = y;
        // cout <<"  " << (long) *pX << " "<<(long) *pY << endl;
        return;
    }
    if (cX < FIELD_WIDTH - WALL_THICKNESS
         && cutVertical (cX, cY, *pX, *pY, FIELD_WIDTH - WALL_THICKNESS, &y) == true) {
        // cout << "Angle Clipping " << (long) *pX << " "<<(long) *pY;
        *pX = FIELD_WIDTH - WALL_THICKNESS;
        *pY = y;
        // cout <<"  " << (long) *pX << " "<<(long) *pY << endl;
        return;
    }
    if (cY > WALL_THICKNESS
         && cutHorizontal (cX, cY, *pX, *pY, WALL_THICKNESS, &x) == true) {
        // cout << "Angle Clipping " << (long) *pX << " "<<(long) *pY;
        *pX = x;
        *pY = WALL_THICKNESS;
        // cout <<"  " << (long) *pX << " "<<(long) *pY << endl;
        return;
    }
    if (cY < FIELD_LENGTH - WALL_THICKNESS
         && cutHorizontal (cX, cY, *pX, *pY, FIELD_LENGTH - WALL_THICKNESS, &x) == true) {
        // cout << "Angle Clipping " << (long) *pX << " "<<(long) *pY;
        *pX = x;
        *pY = FIELD_LENGTH - WALL_THICKNESS;
        // cout <<"  " << (long) *pX << " "<<(long) *pY << endl;
        return;
    }
    angleClippingMore (cX, cY, pX, pY);
}

bool cutVerticalMore (double cX, double cY, double pX, double pY, double Xval, double *resY) {
    if (cX == pX)
        return false;
    if ((Xval - cX) * (Xval - pX) > 0)
        return false;

    *resY = (Xval - cX) * (cY - pY) / (cX - pX) + cY;
    //  cout << " Vertical " << Xval << " " <<pX<<" "<<pY<< " " << *resY << endl;
    if ((*resY - 0.0) * (*resY - FIELD_LENGTH) > 0)
        return false;
    else
        return true;
}

bool cutHorizontalMore (double cX, double cY, double pX, double pY, double Yval, double *resX) {
    if (cY == pY)
        return false;
    if ((Yval - cY) * (Yval - pY) > 0)
        return false;

    *resX = (Yval - cY) * (cX - pX) / (cY - pY) + cX;
    //  cout << " Horizontal " << Yval << " " <<pX<<" "<<pY<< " " 
    //       << *resX << endl;
    if ((*resX - 0.0) * (*resX - FIELD_WIDTH) > 0)
        return false;
    else
        return true;
}

void angleClippingMore (double cX, double cY, double *pX, double *pY) {
    double x, y;
    //  cout << "angleclipping " << cX<<" "<<cY<<" "<<*pX<<" "<<*pY<<endl;
    if (cX > 0.0 && cutVerticalMore (cX, cY, *pX, *pY, 0.0, &y) == true) {

        // cout << "Angle Clipping " << (long) *pX << " "<<(long) *pY;
        *pX = 0.0;
        *pY = y;
        // cout <<"  " << (long) *pX << " "<<(long) *pY << endl;
        return;
    }
    if (cX < FIELD_WIDTH && cutVerticalMore (cX, cY, *pX, *pY, FIELD_WIDTH, &y) == true) {
        // cout << "Angle Clipping " << (long) *pX << " "<<(long) *pY;
        *pX = FIELD_WIDTH;
        *pY = y;
        // cout <<"  " << (long) *pX << " "<<(long) *pY << endl;
        return;
    }
    if (cY > 0.0 && cutHorizontalMore (cX, cY, *pX, *pY, 0.0, &x) == true) {
        // cout << "Angle Clipping " << (long) *pX << " "<<(long) *pY;
        *pX = x;
        *pY = 0.0;
        // cout <<"  " << (long) *pX << " "<<(long) *pY << endl;
        return;
    }
    if (cY < FIELD_LENGTH && cutHorizontalMore (cX, cY, *pX, *pY, FIELD_LENGTH, &x) == true) {
        // cout << "Angle Clipping " << (long) *pX << " "<<(long) *pY;
        *pX = x;
        *pY = FIELD_LENGTH;
        // cout <<"  " << (long) *pX << " "<<(long) *pY << endl;
        return;
    }
}

int ReadSignedInt (void) {
    char str[64];
    cin >> str;
    return atoi (str);
}

double getHeadingBetween (double fromX, double fromY, double toX, double toY) {
    double relx = toX - fromX;
    double rely = toY - fromY;
    if (relx == 0 && rely == 0)
        return 0;
    return atan2 (rely, relx);
}

double getDistanceBetween (double fromX, double fromY, double toX, double toY) {
    double relx = toX - fromX;
    double rely = toY - fromY;
    return sqrt (SQUARE (relx) + SQUARE (rely));
}

void calPosition (double x, double y, double h, double d, double ang, double *pX, double *pY) {

    double deg, rad;

    deg = h + (double) ang;    // angle of object relative to horizontal line of the field
    while (deg >= 360)
        deg -= 360;
    while (deg < 0)
        deg += 360;

    rad = deg / 180.0 * pi;

    *pX = (x + d * cos (rad));
    *pY = (y + d * sin (rad));
    //angleClipping(x,y,pX,pY);
}





