/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: Common.h 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 *
 * contains basic utilities and functions
 * 
**/

#ifndef _Common_h_DEFINED
#define _Common_h_DEFINED

/*
#include <OPENR/OSubject.h>
#include <OPENR/OObserverVector.h>
#include <OPENR/OPENREvent.h>
#include <OPENR/OPENRAPI.h>
*/

#ifdef OFFLINE
    typedef long slongword;
    typedef char byte;
#else // OFFLINE
#include <Types.h>
#endif // OFFLINE
#include <math.h>

#include "RobotDef.h"
#include "FieldDef.h"
#include "BallDef.h"

typedef unsigned char uchar;
typedef unsigned short ushort;
typedef unsigned long ulong;
typedef void * voidPtr;

static const double pi = M_PI;

static const int GLOBAL = 0;
static const int LOCAL = 1;

static const double FULL_CIRCLE = 360;
static const double HALF_CIRCLE = 180;
static const double QUARTER_CIRCLE = 90;

// these variables to avoid computation
static const double DEG_OVER_RAD = 180/M_PI;
static const double RAD_OVER_DEG = M_PI/180;
static const double TWO_PI_INV = 1/(2*M_PI);

template<class T>
inline static T MAX(const T& x, const T& y) {
	return (x > y ? x : y);
}
inline static double MAX(const double x, const int y) {
	return (x > y ? x : y);
}
inline static double MAX(const int x, const double y) {
	return (x > y ? x : y);
}

template<class T>
inline static T MIN(const T& x, const T& y) {
	return (x > y ? y : x);
}
inline static double MIN(const double x, const int y) {
	return (x > y ? y : x);
}
inline static double MIN(const int x, const double y) {
	return (x > y ? y : x);
}

template<class T>
inline static T ABS(const T& x) {
	return (x > 0 ? x : -x);
}

template<class T>
inline static T SQUARE(const T& x) {
	return x*x;
}

inline static int round(const double x) {
	if (x > 0) {
		return (int)((x) + 0.5);
	} else {
		return (int)((x) - 0.5);
	}
	return 0;
}

inline static double RAD2DEG(const double x) { return ((x) / M_PI * 180.0); }
inline static double DEG2RAD(const double x) { return ((x) / 180.0 * M_PI); }
inline static double DEG2RADSQ(const double x) {return (x*M_PI/180*M_PI/180); }

inline static slongword RAD2MICRO(const double x) { return (slongword) ((x) * 1000000); }
inline static slongword DEG2MICRO(const double x) { return RAD2MICRO(DEG2RAD(x)); }
inline static double MICRO2RAD(const slongword x) { return ((x) / 1000000.0); }
inline static double MICRO2DEG(const slongword x) { return RAD2DEG(MICRO2RAD(x)); }

template<class T>
inline static T DISTANCE_SQR(const T& xA, const T& yA, const T& xB, const T& yB) {
	return SQUARE(xA-xB) + SQUARE(yA-yB);
}

inline static double DISTANCE(const double xA, const double yA, const double xB, const double yB) {
	return sqrt(DISTANCE_SQR(xA,yA,xB,yB));
}

template<class T>
inline static int SIGN(const T& x) {
	return (x < 0 ? -1 : 1);
}

template<class T>
inline static T CLIP(const T& x, const T& bound) {    /* clips (+/-) x to bound */
	return ((x < -bound || x > bound) ? bound * SIGN(x) : x);
}

static inline int DECREMENT(int& x) {
	return (x = (x <= 0 ? 0 : x - 1));
}

//Robot dimension constants, etc.

static const int ROBOTS = 8;
static const int NUM_TEAM_MEMBER = 4;
static const int NUM_OPPONENTS = 4;
static const int NUM_FRIENDS_VIS = 3;
static const int NUM_FORWARDS = 2;
static const int VISION_BALL = 0;
static const int WIRELESS_BALL = 1;
static const int GPS_BALL = 2;
static const int TEAMMATE_HAS_BALL = 3; //used by maverick.. shouldn't be here but otherwise the number 3 might get used by someone else

double NormalizeAngle_0_360 (double angle);
double NormalizeAngle_180 (double angle);
double NormalizeAngle_PI (double angle);
double NormalizeAngle_m90_270(double angle);
double NormalizeAngle_90_450 (double angle);
int byte2UInt (byte);
int ReadSignedInt (void);
void FieldClipping (double &x, double &y);
bool IsInField (double x, double y);

/* 
 * @return true if (cX,cY)-(pX,pY) cuts vertical line with 
 * Xval (Xval,WALL_THICKNESS) - (Xval,FIELD_LENGTH - WALL_THICKNESS) at resY
 * otherwise false.
**/
bool cutVertical (double cX, double cY, double pX, double pY, double Xval, double *resY);

/* 
 * @return true if (cX,cY)-(pX,pY) cuts horizontal line with 
 * Yval (WALL_THICKNESS,Yval) - (FIELD_WIDTH-WALL_THICKNESS,Yval) at resX
 * otherwise false.
**/
bool cutHorizontal (double cX, double cY, double pX, double pY, double Yval, double *resX);

/* if (pX,pY) is out of the field, it should be clipped but
 * the gradient of the line (pX,pY) and (cX,cY) should not be modified.
 * So check if (pX,pY)-(cX,cY) cuts the 4 bounderies then that is
 * the clipped point.
 * @param cX is the centered X
**/
void angleClipping (double cX, double cY, double *pX, double *pY);

/* 
 * @return true if (cX,cY)-(pX,pY) cuts vertical line with 
 * Xval (Xval,0) - (Xval,FIELD_LENGTH) at resY
 * otherwise false.
**/
bool cutVerticalMore (double cX, double cY, double pX, double pY, double Xval, double *resY);

/* 
 * @return true if (cX,cY)-(pX,pY) cuts horizontal line with 
 * Yval (0,Yval) - (FIELD_WIDTH,Yval) at resX
 * otherwise false.
**/
bool cutHorizontalMore (double cX, double cY, double pX, double pY, double Yval, double *resX);

/* if (pX,pY) is out of the field, it should be clipped but
 * the gradient of the line (pX,pY) and (cX,cY) should not be modified.
 * So check if (pX,pY)-(cX,cY) cuts the 4 bounderies then that is
 * the clipped point.
 * @param cX is centered X
**/
void angleClippingMore (double cX, double cY, double *pX, double *pY);

inline int determinant (int a1, int b1, int c1, int a2, int b2, int c2, int a3, int b3, int c3) {
    return a1 * b2 * c3 - a1 * b3 * c2 + a2 * b3 * c1 - a2 * b1 * c3 +
           a3 * b1 * c2 - a3 * b2 * c1;
}

/* 
 * @return radian
**/
double getHeadingBetween (double fromX, double fromY, double toX, double toY);

double getDistanceBetween (double fromX, double fromY, double toX, double toY);

/*
 * spits out the 95% confidence factor of the variance of
 * of a given value
**/
inline double get95CF (double standardDeviation) {
  return SQUARE (standardDeviation) / 4;
}


/*
 * calculate the offset position(pX,pY) which are distance d away at an angle ang from the pivot position (x, y, h)
 * @param x the current x loc
 * @param y the current y loc
 * @param h the current heading (absolute)
 * @param d the distance to walk
 * @param ang the angle to walk, relative to the current heading (ccw is +ve)
 * @image html calPosition.gif
**/
void calPosition (double x, double y, double h, double d, double ang, double *pX, double *pY);


//====================================================================
//quicksort template function
//====================================================================
struct sortee {
	double id;
	double weight;
    bool opponent;

	bool operator>=(sortee s) {
		return this->weight>=s.weight;
	}

	bool operator<=(sortee s) {
		return this->weight<=s.weight;
	}
};

template<class T>
void swap(int i, int j, T *array) {
   T temp = array[i];
   array[i] = array[j];
   array[j] = temp;
}


template<class T>
int partition(int first, int last, T *array) {

   int left  = first;
   int right = last;

   while (left < right) {
      while ((array[left] <= array[first]) && (left<last)) {
         left++;
      }
      while ((array[right] >= array[first]) && (right>first)) {
         right--;
      }
      if (left < right)
         swap(left, right, array);
   }
   swap(right, first, array);
   return right;
}


template<class T>
void qSort (int first, int last, T *array) {
   int middle;
   if (first < last){ 
       middle = partition(first,last,array);
       qSort(first, middle-1, array);
       qSort(middle+1, last, array);
   }
}

template<class T>
void quickSort(T *array, int arraySize) {
   int last = arraySize - 1;
   qSort (0,last,array);
}
//====================================================================


#endif // _Common_h_DEFINED


