/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * UNSW 2003 Robocup (Nicodemus Sutanto)
 *
 * Last modification background information
 * $Id: SharedMemoryDef.h 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
**/

#ifndef OFFLINE
#include <OPENR/OObject.h>
#endif // OFFLINE

#ifndef SharedMemoryDef_h_DEFINED
#define SharedMemoryDef_h_DEFINED

#include "SwitchBoard.h"
#include "VisionDef.h"
#include "PackageDef.h"
#include "Common.h"
#include "WirelessSwitchboard.h"


// For inter-robot latency testing
static const int NORMAL_WM = 0;
static const int INIT_LAT_WM = -1;



// used to store information form kalman filter updates for sending to teammates
// uses floats to save space, floats are 4 bytes, doubles are 8 bytes
struct IFKUpdateData {

    float iVecx;
    float iVecy;
    
    float iMat11;
    float iMat12;
    float iMat21;
    float iMat22;

};

// The information to be passed when sharing world model
// with other robots
struct WMShareObj {
    //WMObj obj;
	float posx, posy;
	float posh;
	float posVar, hVar;

	unsigned int behavioursVal;
	//team maverick stuff
	/*
	unsigned int ballDist:29;
	unsigned int amThirdPlayer:1;
	unsigned int playerType:2;
	*/
	
    double ballx;
    double bally;
    double ballvar;

    // int teamCForwardRole;
    
	// information for information form kalman filter tracking opponents
    IFKUpdateData opponentsUpdate[NUM_OPPONENTS];
	
#ifdef IR_LATENCY_TEST
	// will be NORMAL_WM, INIT_LAT_WM, or the player number to the robot
	// we are responding to for latency testing
	int latMode;
	SystemTime waited;
#endif // IR_LATENCY_TEST

};

struct WMShareInfo {
    
        WMShareObj obj;
        int playerNum;
    
};

#ifdef DELAY_TESTING
static const bool isDelayTesting = true;
#else
static const bool isDelayTesting = false;
#endif

#ifdef SENDING_DEBUG

// Debug information to be sent to Robocommander
struct RC_DEBUG_DATA {
    
    char message[100];

#ifdef SENDING_WORLD_MODEL_DEBUG
    char world_model_debug_data[560];
#endif // SENDING_WORLD_MODEL_DEBUG

#ifdef SENDING_CPLANE
#ifdef DELAY_TESTING
char cplane_data[INT_BYTES + COMPRESSED_CPLANE_SIZE + SENSOR_SIZE + SANITY_SIZE + OBJ_SIZE + GPS_SIZE + 4]; // for counter
#else
char cplane_data[INT_BYTES + COMPRESSED_CPLANE_SIZE + SENSOR_SIZE + SANITY_SIZE + OBJ_SIZE + GPS_SIZE];
#endif // DELAY_TESTING
#endif // SENDING_CPLANE

#ifdef SENDING_OPLANE
    char oplane_data[OBJ_SIZE];
#endif // SENDING_OPLANE

#ifdef SENDING_YUVPLANE
    char yuvplane_data[YUVPLANE_SIZE];
    char yuvCplane_data[CPLANE_SIZE];
#endif // SENDING_YUVPLANE

#ifdef SENDING_LINE_DEBUG
    char line_debug_data[CPLANE_SIZE * DEBUG_PLANES];
#endif // SENDING_RSENDING_LINE_DEBUGAYMONDS_CRAP

};

#endif // SENDING_DEBUG

#endif // SharedMemoryDef_h_DEFINED
